##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2019 Vesa-Pekka Palmu <vpalmu@depili.fi>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
##

import sigrokdecode as srd
from math import ceil
from .lists import *

L = len(cmds)
RX = 0
TX = 1

# Don't forget to keep this in sync with 'cmds' is lists.py.
class Ann:
    PAGE, GBV, GWV, GSV, GLV, GRPC, SBV, SWV, SSV, RPC, LINE, RECT, FRECT, \
    PIXEL, GBVA, GWVA, SBVA, GBVR, GWVR, GSVR, GLVR, GRPCR, SBVR, SWVR, SSVR, \
    RPCR, LINER, RECTR, FRECTR, PIXELR, GBVAR, GWVAR, SBVAR, ACK, NACK, SWVA, \
    SWVAR, GCV, GCVR, SCV, SCVR, BIT, FIELD, WARN = range(L + 3)

def cmd_annotation_classes():
    return tuple([tuple([cmd[0].lower(), cmd[1]]) for cmd in cmds.values()])

class Decoder(srd.Decoder):
    api_version = 3
    id = 'amulet_ascii'
    name = 'Amulet ASCII'
    longname = 'Amulet LCD ASCII'
    desc = 'Amulet Technologies LCD controller ASCII protocol.'
    license = 'gplv3+'
    inputs = ['uart']
    outputs = []
    tags = ['Display']
    annotations = cmd_annotation_classes() + (
        ('bit', 'Bit'),
        ('field', 'Field'),
        ('warning', 'Warning'),
    )
    annotation_rows = (
        ('bits', 'Bits', (L + 0,)),
        ('fields', 'Fields', (L + 1,)),
        ('commands', 'Commands', tuple(range(len(cmds)))),
        ('warnings', 'Warnings', (L + 2,)),
    )
    options = (
        {'id': 'ms_chan', 'desc': 'Master -> slave channel',
            'default': 'RX', 'values': ('RX', 'TX')},
        {'id': 'sm_chan', 'desc': 'Slave -> master channel',
            'default': 'TX', 'values': ('RX', 'TX')},
    )

    def __init__(self):
        self.reset()

    def reset(self):
        self.state = None
        self.cmdstate = None

        # Build dict mapping command keys to handler functions. Each
        # command in 'cmds' (defined in lists.py) has a matching
        # handler self.handle_<shortname>.
        def get_handler(cmd):
            s = 'handle_%s' % cmds[cmd][0].lower().replace('/', '_')
            return getattr(self, s)
        self.cmd_handlers = dict((cmd, get_handler(cmd)) for cmd in cmds.keys())

    def start(self):
        self.out_ann = self.register(srd.OUTPUT_ANN)

    def putx(self, data):
        # Simplification, most annotations span exactly one SPI byte/packet.
        self.put(self.ss, self.es, self.out_ann, data)

    def putf(self, data):
        self.put(self.ss_field, self.es_field, self.out_ann, data)

    def putc(self, data):
        self.put(self.ss_cmd, self.es_cmd, self.out_ann, data)

    def cmd_ann_list(self):
        x, s = cmds[self.state][0], cmds[self.state][1]
        return ['Command: %s (%s)' % (s, x), 'Command: %s' % s,
                'Cmd: %s' % s, 'Cmd: %s' % x, x]

    def emit_cmd_byte(self):
        self.ss_cmd = self.ss
        self.putx([Ann.FIELD, self.cmd_ann_list()])

    def emit_addr_bytes(self, pdata):
        if self.cmdstate == 2:
            self.ss_field = self.ss
            self.addr = chr(pdata)
            self.putx([Ann.BIT, ['Address high nibble: %c' % pdata,
                 'Addr high 0x%c' % pdata, 'Addr h 0x%c' % pdata]])
        elif self.cmdstate == 3:
            self.es_field = self.es
            self.addr += chr(pdata)
            self.addr = int(self.addr, 16)
            self.putx([Ann.BIT, ['Address low nibble: %c' % pdata,
                 'Addr low 0x%c' % pdata, 'Addr l 0x%c' % pdata]])
            self.putf([Ann.FIELD, ['Address: 0x%02X' % self.addr,
                'Addr: 0x%02X' % self.addr, '0x%02X' % self.addr]])

    def emit_cmd_end(self, data):
        self.es_cmd = self.es
        self.putc(data)
        self.state = None

    def handle_read(self, data):
        if self.cmdstate == 1:
            self.emit_cmd_byte()
            self.addr = 0
        elif self.cmdstate == 2:
            self.emit_addr_bytes(pdata)
        elif self.cmdstate == 3:
            self.emit_addr_bytes(pdata)
        self.cmdstate += 1

    def handle_set_common(self, pdata):
        if self.cmdstate == 1:
            self.addr = 0
        self.emit_addr_bytes(pdata)

    def emit_not_implemented(self, data):
        self.es_cmd = self.es
        self.putc([Ann.WARN, ['Command not decoded', 'Not decoded']])
        self.emit_cmd_end(data)

    def handle_string(self, pdata, ann_class):
        # TODO: unicode / string modifiers...
        self.handle_set_common(pdata)
        if self.cmdstate == 4:
            self.ss_field = self.ss
            self.value = ''
        if pdata == 0x00:
            # Null terminated string ends.
            self.es_field = self.es
            self.putx([Ann.BIT, ['NULL']])
            self.putf([Ann.FIELD, ['Value: %s' % self.value,
                'Val: %s' % self.value, '%s' % self.value]])
            self.emit_cmd_end([ann_class, self.cmd_ann_list()])
            return
        if self.cmdstate > 3:
            self.value += chr(pdata)
            self.putx([Ann.BIT, ['%c' % pdata]])
        self.cmdstate += 1

    # Command handlers

    # Page change 0xA0, 0x02, index_high, index_low, checksum
    def handle_page(self, pdata):
        if self.cmdstate == 2:
            if pdata == 0x02:
                self.ss_field = self.ss_cmd
                self.es_field = self.es
                self.putf([Ann.FIELD, self.cmd_ann_list()])
                self.checksum = 0xA0 + 0x02
            else:
                self.putx([Ann.WARN, ['Illegal second byte for page change',
                                      'Illegal byte']])
                self.state = None
        elif self.cmdstate == 3:
            self.ss_field = self.ss
            self.checksum += pdata
            self.page[0] = pdata
        elif self.cmdstate == 4:
            self.checksum += pdata
            self.page[1] = pdata
            self.es_field = self.es
            if self.page[0] == self.page [1] == 0xFF:
                # Soft reset trigger
                self.putf(Ann.WARN, ['Soft reset', 'Reset'])
            else:
                page = chr(self.page[0]) + chr(self.page[1])
                self.putf(Ann.FIELD, ['Page index: 0x%s' % page,
                                      'Page: 0x%s' % page, '0x%s' % page])
        elif self.cmdstate == 5:
            self.checksum += pdata
            if (self.checksum & 0xFF) != 0:
                self.putx([Ann.WARN, ['Checksum error', 'Error', 'ERR']])
            else:
                self.putx([Ann.FIELD, ['Checksum OK', 'OK']])
            self.emit_cmd_end(Ann.PAGE)
        self.cmdstate += 1

    # Value reads: command byte, address high nibble, address low nibble

    # Get byte value
    def handle_gbv(self, pdata):
        self.handle_read(pdata)
        self.emit_cmd_end([Ann.GBV, self.cmd_ann_list()])

    # Get word value
    def handle_gwv(self, pdata):
        self.handle_read(pdata)
        self.emit_cmd_end([Ann.GWV, self.cmd_ann_list()])

    # Get string value
    def handle_gsv(self, pdata):
        self.handle_read(pdata)
        self.emit_cmd_end([Ann.GSV, self.cmd_ann_list()])

    # Get label value
    def handle_glv(self, pdata):
        self.handle_read(pdata)
        self.emit_cmd_end([Ann.GLV, self.cmd_ann_list()])

    # Get RPC buffer
    def handle_grpc(self, pdata):
        if self.cmdstate == 2:
            self.ss_field = self.ss
            self.flags = int(chr(pdata), 16) << 4
        elif self.cmdstate == 3:
            self.flags += int(chr(pdata), 16)
            self.es_field = self.es
            self.putf([Ann.FIELD, ['RPC flag: 0x%02X' % self.flags]])
            self.emit_cmd_end([Ann.GRPC, self.cmd_ann_list()])

    # Get byte value array
    def handle_gbva(self, pdata):
        self.handle_read(pdata)
        self.emit_cmd_end([Ann.GBVA, self.cmd_ann_list()])

    # Get word value array
    def handle_gwva(self, pdata):
        self.handle_read(pdata)
        self.emit_cmd_end([Ann.GWVA, self.cmd_ann_list()])

    # Get color variable
    def handle_gcv(self, pdata):
        self.handle_read(pdata)
        self.emit_cmd_end([Ann.GCV, self.cmd_ann_list()])

    # Value setters: command byte, address high nibble, address low nibble, data bytes

    # Set byte value data = high nibble, low nibble
    def handle_sbv(self, pdata):
        self.handle_set_common(pdata)
        if self.cmdstate == 4:
            self.ss_field = self.ss
            self.value = chr(pdata)
        elif self.cmdstate == 5:
            self.value += chr(pdata)
            self.es_field = self.es
            self.putf([Ann.FIELD, ['Value: 0x%s' % self.value,
                'Val: 0x%s' % self.value, '0x%s' % self.value]])
            self.emit_cmd_end([Ann.SBV, self.cmd_ann_list()])
        self.cmdstate += 1

    # Set word value, msb high, msb low, lsb high, lsb low
    def handle_swv(self, pdata):
        self.handle_set_common(pdata)
        if self.cmdstate > 3:
            nibble = self.cmdstate - 4
            if nibble == 0:
                self.ss_field = self.ss
                self.value = 0
            self.value += int(chr(pdata), 16) << 12 - (4 * nibble)
            if nibble == 3:
                self.es_field = self.es
                self.putf([Ann.FIELD, ['Value: 0x%04x' % self.value,
                    'Val: 0x%04x' % self.value, '0x%04x' % self.value]])
                self.emit_cmd_end([Ann.SWV, self.cmd_ann_list()])
                return
        self.cmdstate += 1

    # Set string value, null terminated utf8 strings
    def handle_ssv(self, pdata):
        self.handle_string(pdata, Ann.SSV)

    # Set byte value array
    def handle_sbva(self, pdata):
        nibble = (self.cmdstate - 3) % 2
        if self.cmdstate == 2:
            self.addr = int(chr(pdata), 16) << 4
            self.ss_field = self.ss
            self.putx([Ann.BIT, ['Address high nibble: %c' % pdata,
                 'Addr high 0x%c' % pdata, '0x%c' % pdata]])
        elif self.cmdstate == 3:
            self.addr += int(chr(pdata), 16)
            self.es_field = self.ss
            self.putx([Ann.BIT, ['Address low nibble: %c' % pdata,
                 'Addr low 0x%c' % pdata, '0x%c' % pdata]])
            self.putf([Ann.FIELD, ['Address: 0x%02X' % self.addr,
                'Addr: 0x%02X' % self.addr, '0x%02X' % self.addr]])
        elif stage == 2:
            if pdata == 0x00:
                # Null terminated list
                self.emit_cmd_end([Ann.SBVA, self.cmd_ann_list()])
                return
            self.value = int(chr(pdata), 16) << 4
        else:
            self.value += int(chr(pdata), 16)
            self.es_field = self.es
            self.putf([Ann.FIELD, ['Value 0x%02X' % self.value,
                                   '0x%02X' % self.value]])
        self.cmdstate += 1

    # Set word value array
    def handle_swva(self, pdata):
        nibble = (self.cmdstate - 3) % 4
        if self.cmdstate == 2:
            self.addr = int(chr(pdata), 16) << 4
            self.ss_field = self.ss
            self.putx([Ann.BIT, ['Address high nibble: %c' % pdata,
                  'Addr high 0x%c' % pdata, '0x%c' % pdata]])
        elif self.cmdstate == 3:
            self.addr += int(chr(pdata), 16)
            self.es_field = self.ss
            self.putx([Ann.BIT, ['Address low nibble: %c' % pdata,
                 'Addr low 0x%c' % pdata, '0x%c' % pdata]])
            self.putf([Ann.FIELD, ['Address: 0x%02X' % self.addr,
                 'Addr: 0x%02X' % self.addr, '0x%02X' % self.addr]])
            self.value = 0
        else:
            self.value += int(chr(pdata), 16) << 12 - (4 * nibble)
            if nibble == 0:
                if pdata == 0x00:
                    # Null terminated list
                    self.emit_cmd_end([Ann.SWVA, self.cmd_ann_list()])
                    return
                self.ss_field = self.ss
            if nibble == 3:
                self.es_field = self.es
                self.putf([Ann.FIELD, ['Value 0x%04X' % self.value,
                                       '0x%04X' % self.value]])
                self.cmdstate += 1

    # Set color variable
    def handle_scv(self, pdata):
        if self.cmdstate == 8:
            self.emit_not_implemented([Ann.SCV, self.cmd_ann_list()])
        self.cmdstate += 1

    # RPC trigger
    def handle_rpc(self, pdata):
        self.handle_read(pdata)
        self.emit_cmd_end([Ann.RPC, self.cmd_ann_list()])

    # Drawing

    # Decode pair of (x,y) 16bit coordinates
    def decode_coords(self, pdata):
        if self.cmdstate == 1:
            self.coords[0] = 0
            self.coords[1] = 0
            self.coords[2] = 0
            self.coords[3] = 0
        if self.cmdstate < 18:
            # Coordinates
            nibble = (self.cmdstate - 1) % 4
            i = (self.cmdstate - 1) / 4
            self.coords[i] += int(chr(pdata), 16) << 12 - (4 * nibble)
            if nibble == 0:
                self.ss_field = self.ss
            elif nibble == 3:
                self.es_field = self.es
                self.putf([Ann.FIELD, ['Coordinate 0x%04X' % self.coords[i]],
                                      ['0x%04X' % self.coords[i]]])

    # TODO: There are actually two protocol revisions for drawing.
    # Both use 4 bytes for 16bit x and y pairs for start and end.
    # The older follows this by a pattern selector and then line weight.
    # Newer version has 6 bytes for 8bit RGB color...

    # Draw line
    def handle_line(self, pdata):
        decode_coords(pdata)
        if self.cmdstate == 18:
            self.es_cmd = self.es
            self.putc([Ann.LINE, self.cmd_ann_list()])
            self.putc([Ann.WARN, ['Line pattern / Color not implemented']])
            self.state = None
        self.cmdstate += 1

    # Draw rectange
    def handle_rect(self, pdata):
        decode_coords(pdata)
        if self.cmdstate == 18:
            self.es_cmd = self.es
            self.putc([Ann.RECT, self.cmd_ann_list()])
            self.putc([Ann.WARN, ['Line pattern / Color not implemented']])
            self.state = None
        self.cmdstate += 1

    # Draw filled rectangle
    def handle_frect(self, pdata):
        decode_coords(pdata)
        if self.cmdstate == 18:
            self.es_cmd = self.es
            self.putc([Ann.FRECT, self.cmd_ann_list()])
            self.putc([Ann.WARN, ['Fill pattern / Color not implemented']])
            self.state = None
        self.cmdstate += 1

    # Draw pixel
    def handle_pixel(self, pdata):
        self.es_cmd = self.es
        self.putc([Ann.WARN, ['Draw pixel documentation is missing.', 'Undocumented']])
        self.state = None

    # Replies
    def handle_gbvr(self, pdata):
        self.emit_add_bytes(pdata)
        if self.cmdstate == 4:
            self.ss_field = self.ss
            self.value = int(chr(pdata), 16) << 4
            self.putx([Ann.BIT, ['High nibble 0x%s' % pdata, '0x%s' % pdata]])
        elif self.cmdstate == 5:
            self.value += int(chr(pdata), 16)
            self.putx([Ann.BIT, ['Low nibble 0x%s' % pdata, '0x%s' % pdata]])
            self.es_field = self.es
            self.putf([Ann.FIELD, ['Value: 0x%02X' % self.value,
                                   '0x%02X' % self.value]])
            self.emit_cmd_end([Ann.GBVR, self.cmd_ann_list()])
        self.cmdstate += 1

    def handle_gwvr(self, pdata):
        self.emit_add_bytes(pdata)
        if self.cmdstate > 3:
            nibble = self.cmdstate - 3
            if nibble == 0:
                self.value = 0
                self.ss_field = self.ss
            self.value += int(chr(pdata), 16) << 12 - (4 * nibble)
            self.putx([Ann.BIT, ['0x%s' % pdata]])
            if nibble == 3:
                self.putf([Ann.FIELD, ['Value: 0x%04x' % self.value,
                                       '0x%04X' % self.value]])
                self.es_cmd = self.ss
                self.emit_cmd_end([Ann.GWVR, self.cmd_ann_list()])
        self.cmdstate += 1

    def handle_gsvr(self, pdata):
        self.handle_string(pdata, Ann.GSVR)

    def handle_glvr(self, pdata):
        self.handle_string(pdata, Ann.GLVR)

    def handle_grpcr(self, pdata):
        self.handle_addr(pdata)
        if self.cmdstate > 3:
            nibble = (self.cmdstate - 3) % 2
            if nibble == 0:
                if pdata == 0x00:
                    self.emit_cmd_end([Ann.GRPCR, self.cmd_ann_list()])
                    return
                self.value = int(chr(pdata), 16) << 4
                self.ss_field = self.ss
                self.putx([Ann.BIT, ['0x%s' % pdata]])
            if nibble == 2:
                self.value += int(chr(pdata), 16)
                self.es_field = self.es
                self.putx([Ann.BIT, ['0x%s' % pdata]])
                self.putf([Ann.FIELD, ['0x%02X' % self.value]])
        self.cmdstate += 1

    def handle_sbvr(self, pdata):
        self.handle_set_common(pdata)
        if self.cmdstate == 4:
            self.ss_field = self.ss
            self.value = chr(pdata)
        elif self.cmdstate == 5:
            self.value += chr(pdata)
            self.es_field = self.es
            self.putf([Ann.FIELD, ['Value: 0x%s' % self.value,
                'Val: 0x%s' % self.value, '0x%s' % self.value]])
            self.emit_cmd_end([Ann.SBVR, self.cmd_ann_list()])
        self.cmdstate += 1

    def handle_swvr(self, pdata):
        self.handle_set_common(pdata)
        if self.cmdstate == 4:
            self.ss_field = self.ss
            self.value = (pdata - 0x30) << 4
        elif self.cmdstate == 5:
            self.value += (pdata - 0x30)
            self.value = self.value << 8
        elif self.cmdstate == 6:
            self.value += (pdata - 0x30) << 4
        elif self.cmdstate == 7:
            self.value += (pdata - 0x30)
            self.es_field = self.es
            self.putf([Ann.FIELD, ['Value: 0x%04x' % self.value,
                'Val: 0x%04x' % self.value, '0x%04x' % self.value]])
            self.emit_cmd_end([Ann.SWVR, self.cmd_ann_list()])
            self.state = None
        self.cmdstate += 1

    def handle_ssvr(self, pdata):
        self.handle_string(pdata, Ann.SSVR)

    def handle_rpcr(self, pdata):
        self.handle_read(pdata)
        self.emit_cmd_end([Ann.RPCR, self.cmd_ann_list()])

    def handle_liner(self, pdata):
        decode_coords(pdata)
        if self.cmdstate == 18:
            self.es_cmd = self.es
            self.putc([Ann.LINER, self.cmd_ann_list()])
            self.putc([Ann.WARN, ['Line pattern / Color not implemented']])
            self.state = None
        self.cmdstate += 1

    def handle_rectr(self, pdata):
        decode_coords(pdata)
        if self.cmdstate == 18:
            self.es_cmd = self.es
            self.putc([Ann.RECTR, self.cmd_ann_list()])
            self.putc([Ann.WARN, ['Line pattern / Color not implemented']])
            self.state = None
        self.cmdstate += 1

    def handle_frectr(self, pdata):
        decode_coords(pdata)
        if self.cmdstate == 18:
            self.es_cmd = self.es
            self.putc([Ann.FRECTR, self.cmd_ann_list()])
            self.putc([Ann.WARN, ['Line pattern / Color not implemented']])
            self.state = None
        self.cmdstate += 1

    def handle_pixelr(self, pdata):
        self.es_cmd = self.es
        self.putc([Ann.WARN,['Draw pixel documentation is missing.', 'Undocumented']])
        self.state = None

    def handle_gbvar(self, pdata):
        nibble = (self.cmdstate - 3) % 2
        if self.cmdstate == 2:
            self.addr = int(chr(pdata), 16) << 4
            self.ss_field = self.ss
            self.putx([Ann.BIT, ['Address high nibble: %c' % pdata,
                 'Addr high 0x%c' % pdata, '0x%c' % pdata]])
        elif self.cmdstate == 3:
            self.addr += int(chr(pdata), 16)
            self.es_field = self.ss
            self.putx([Ann.BIT, ['Address low nibble: %c' % pdata,
                 'Addr low 0x%c' % pdata, '0x%c' % pdata]])
            self.putf([Ann.FIELD, ['Address: 0x%02X' % self.addr,
                'Addr: 0x%02X' % self.addr, '0x%02X' % self.addr]])
        elif stage == 2:
            if pdata == 0x00:
                # Null terminated list
                self.emit_cmd_end([Ann.GBVAR, self.cmd_ann_list()])
                return
            self.value = int(chr(pdata), 16) << 4
        else:
            self.value += int(chr(pdata), 16)
            self.es_field = self.es
            self.putf([Ann.FIELD, ['Value 0x%02X' % self.value,
                                   '0x%02X' % self.value]])
        self.cmdstate += 1

    def handle_gwvar(self, pdata):
        nibble = (self.cmdstate - 3) % 4
        if self.cmdstate == 2:
            self.addr = int(chr(pdata), 16) << 4
            self.ss_field = self.ss
            self.putx([Ann.BIT, ['Address high nibble: %c' % pdata,
                  'Addr high 0x%c' % pdata, '0x%c' % pdata]])
        elif self.cmdstate == 3:
            self.addr += int(chr(pdata), 16)
            self.es_field = self.ss
            self.putx([Ann.BIT, ['Address low nibble: %c' % pdata,
                 'Addr low 0x%c' % pdata, '0x%c' % pdata]])
            self.putf([Ann.FIELD, ['Address: 0x%02X' % self.addr,
                 'Addr: 0x%02X' % self.addr, '0x%02X' % self.addr]])
            self.value = 0
        else:
            self.value += int(chr(pdata), 16) << 12 - (4 * nibble)
            if nibble == 0:
                if pdata == 0x00:
                    # Null terminated list
                    self.emit_cmd_end([Ann.GWVAR, self.cmd_ann_list()])
                    return
                self.ss_field = self.ss
            if nibble == 3:
                self.es_field = self.es
                self.putf([Ann.FIELD, ['Value 0x%04X' % self.value,
                                       '0x%04X' % self.value]])
                self.cmdstate += 1

    # Get byte variable array reply
    def handle_sbvar(self, pdata):
        nibble = (self.cmdstate - 3) % 2
        if self.cmdstate == 2:
            self.addr = int(chr(pdata), 16) << 4
            self.ss_field = self.ss
            self.putx([Ann.BIT, ['Address high nibble: %c' % pdata,
                 'Addr high 0x%c' % pdata, '0x%c' % pdata]])
        elif self.cmdstate == 3:
            self.addr += int(chr(pdata), 16)
            self.es_field = self.ss
            self.putx([Ann.BIT, ['Address low nibble: %c' % pdata,
                 'Addr low 0x%c' % pdata, '0x%c' % pdata]])
            self.putf([Ann.FIELD, ['Address: 0x%02X' % self.addr,
                'Addr: 0x%02X' % self.addr, '0x%02X' % self.addr]])
        elif stage == 2:
            if pdata == 0x00:
                # Null terminated list
                self.emit_cmd_end([Ann.SBVAR, self.cmd_ann_list()])
                return
            self.value = int(chr(pdata), 16) << 4
        else:
            self.value += int(chr(pdata), 16)
            self.es_field = self.es
            self.putf([Ann.FIELD, ['Value 0x%02X' % self.value,
                                   '0x%02X' % self.value]])
        self.cmdstate += 1

    # Set word variable array reply
    def handle_swvar(self, pdata):
        nibble = (self.cmdstate - 3) % 4
        if self.cmdstate == 2:
            self.addr = int(chr(pdata), 16) << 4
            self.ss_field = self.ss
            self.putx([Ann.BIT, ['Address high nibble: %c' % pdata,
                  'Addr high 0x%c' % pdata, '0x%c' % pdata]])
        elif self.cmdstate == 3:
            self.addr += int(chr(pdata), 16)
            self.es_field = self.ss
            self.putx([Ann.BIT, ['Address low nibble: %c' % pdata,
                 'Addr low 0x%c' % pdata, '0x%c' % pdata]])
            self.putf([Ann.FIELD, ['Address: 0x%02X' % self.addr,
                 'Addr: 0x%02X' % self.addr, '0x%02X' % self.addr]])
            self.value = 0
        else:
            self.value += int(chr(pdata), 16) << 12 - (4 * nibble)
            if nibble == 0:
                if pdata == 0x00:
                    # Null terminated list
                    self.emit_cmd_end([Ann.SWVAR, self.cmd_ann_list()])
                    return
                self.ss_field = self.ss
            if nibble == 3:
                self.es_field = self.es
                self.putf([Ann.FIELD, ['Value 0x%04X' % self.value,
                                       '0x%04X' % self.value]])
                self.cmdstate += 1

    def handle_gcvr(self, pdata):
        if self.cmdstate == 8:
            self.emit_not_implemented([Ann.SCV, self.cmd_ann_list()])
        self.cmdstate += 1

    def handle_scvr(self, pdata):
        if self.cmdstate == 8:
            self.emit_not_implemented([Ann.SCV, self.cmd_ann_list()])
        self.cmdstate += 1

    # ACK & NACK

    def handle_ack(self, pdata):
        self.putx([Ann.ACK, self.cmd_ann_list()])
        self.state = None

    def handle_nack(self, pdata):
        self.putx([Ann.NACK, self.cmd_ann_list()])
        self.state = None

    def decode(self, ss, es, data):
        ptype, rxtx, pdata = data

        self.ss, self.es = ss, es

        if ptype != 'DATA':
            return

        # Handle commands.
        try:
            abort_current = (0xD0 <= pdata[0] <= 0xF7) and \
                (not (self.state in cmds_with_high_bytes)) and \
                self.state != None
            if abort_current:
                self.putx([Ann.WARN, ['Command aborted by invalid byte', 'Abort']])
                self.state = pdata[0]
                self.emit_cmd_byte()
                self.cmdstate = 1
            if self.state is None:
                self.state = pdata[0]
                self.emit_cmd_byte()
                self.cmdstate = 1
            self.cmd_handlers[self.state](pdata[0])
        except KeyError:
            self.putx([Ann.WARN, ['Unknown command: 0x%02x' % pdata[0]]])
            self.state = None



##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2019 Jiahao Li <reg@ljh.me>
##
## Permission is hereby granted, free of charge, to any person obtaining a copy
## of this software and associated documentation files (the "Software"), to deal
## in the Software without restriction, including without limitation the rights
## to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
## copies of the Software, and to permit persons to whom the Software is
## furnished to do so, subject to the following conditions:
##
## The above copyright notice and this permission notice shall be included in all
## copies or substantial portions of the Software.
##
## THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
## IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
## FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
## AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
## LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
## OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
## SOFTWARE.

'''
This decoder stacks on top of the 'spi' PD and decodes the protocol spoken
by the Microchip ENC28J60 Ethernet chip.

Details:
http://ww1.microchip.com/downloads/en/DeviceDoc/39662e.pdf
'''

from .pd import Decoder



##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2017 Karl Palsson <karlp@etactica.com>
##
## Permission is hereby granted, free of charge, to any person obtaining a copy
## of this software and associated documentation files (the "Software"), to deal
## in the Software without restriction, including without limitation the rights
## to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
## copies of the Software, and to permit persons to whom the Software is
## furnished to do so, subject to the following conditions:
##
## The above copyright notice and this permission notice shall be included in all
## copies or substantial portions of the Software.
##
## THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
## IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
## FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
## AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
## LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
## OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
## SOFTWARE.

import math
import sigrokdecode as srd
from .lists import *

class Decoder(srd.Decoder):
    api_version = 3
    id = 'ade77xx'
    name = 'ADE77xx'
    longname = 'Analog Devices ADE77xx'
    desc = 'Poly phase multifunction energy metering IC protocol.'
    license = 'mit'
    inputs = ['spi']
    outputs = []
    tags = ['Analog/digital', 'IC', 'Sensor']
    annotations = (
        ('read', 'Register read commands'),
        ('write', 'Register write commands'),
        ('warning', 'Warnings'),
    )
    annotation_rows = (
        ('read', 'Read', (0,)),
        ('write', 'Write', (1,)),
        ('warnings', 'Warnings', (2,)),
    )

    def reset_data(self):
        self.expected = 0
        self.mosi_bytes, self.miso_bytes = [], []

    def __init__(self):
        self.reset()

    def reset(self):
        self.ss_cmd, self.es_cmd = 0, 0
        self.reset_data()

    def start(self):
        self.out_ann = self.register(srd.OUTPUT_ANN)

    def putx(self, data):
        self.put(self.ss_cmd, self.es_cmd, self.out_ann, data)

    def put_warn(self, pos, msg):
        self.put(pos[0], pos[1], self.out_ann, [2, [msg]])

    def decode(self, ss, es, data):
        ptype = data[0]
        if ptype == 'CS-CHANGE':
            # Bear in mind, that CS is optional according to the datasheet.
            # If we transition high mid-stream, toss out our data and restart.
            cs_old, cs_new = data[1:]
            if cs_old is not None and cs_old == 0 and cs_new == 1:
                if len(self.mosi_bytes) > 0 and len(self.mosi_bytes[1:]) < self.expected:
                    # Mark short read/write for reg at least!
                    self.es_cmd = es
                    write, reg = self.cmd & 0x80, self.cmd & 0x7f
                    rblob = regs.get(reg)
                    idx = 1 if write else 0
                    self.putx([idx, ['%s: %s' % (rblob[0], "SHORT")]])
                    self.put_warn([self.ss_cmd, es], "Short transfer!")
                self.reset_data()
            return

        # Don't care about anything else.
        if ptype != 'DATA':
            return
        mosi, miso = data[1:]

        if len(self.mosi_bytes) == 0:
            self.ss_cmd = ss
        self.mosi_bytes.append(mosi)
        self.miso_bytes.append(miso)

        # A transfer is 2-4 bytes, (command + 1..3 byte reg).
        if len(self.mosi_bytes) < 2:
            return

        self.cmd = self.mosi_bytes[0]
        write, reg = self.cmd & 0x80, self.cmd & 0x7f
        rblob = regs.get(reg)
        if not rblob:
            # If you don't have CS, this will _destroy_ comms!
            self.put_warn([self.ss_cmd, es], 'Unknown register!')
            return

        self.expected = math.ceil(rblob[3] / 8)
        if len(self.mosi_bytes[1:]) != self.expected:
            return
        valo, vali = None, None
        self.es_cmd = es
        if self.expected == 3:
            valo = self.mosi_bytes[1] << 16 | self.mosi_bytes[2] << 8 | \
                   self.mosi_bytes[3]
            vali = self.miso_bytes[1] << 16 | self.miso_bytes[2] << 8 | \
                   self.miso_bytes[3]
        elif self.expected == 2:
            valo = self.mosi_bytes[1] << 8 | self.mosi_bytes[2]
            vali = self.miso_bytes[1] << 8 | self.miso_bytes[2]
        elif self.expected == 1:
            valo = self.mosi_bytes[1]
            vali = self.miso_bytes[1]

        if write:
            self.putx([1, ['%s: %#x' % (rblob[0], valo)]])
        else:
            self.putx([0, ['%s: %#x' % (rblob[0], vali)]])

        self.reset_data()



##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2019 Ben Dooks <ben.dooks@codethink.co.uk>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
##

import sigrokdecode as srd

MAX_CHANNELS = 8

class Decoder(srd.Decoder):
    api_version = 3
    id = 'tdm_audio'
    name = 'TDM audio'
    longname = 'Time division multiplex audio'
    desc = 'TDM multi-channel audio protocol.'
    license = 'gplv2+'
    inputs = ['logic']
    outputs = []
    tags = ['Audio']
    channels = (
        { 'id': 'clock', 'name': 'Bitclk', 'desc': 'Data bit clock' },
        { 'id': 'frame', 'name': 'Framesync', 'desc': 'Frame sync' },
        { 'id': 'data', 'name': 'Data', 'desc': 'Serial data' },
    )
    options = (
        {'id': 'bps', 'desc': 'Bits per sample', 'default': 16 },
        {'id': 'channels', 'desc': 'Channels per frame', 'default': MAX_CHANNELS },
        {'id': 'edge', 'desc': 'Clock edge to sample on', 'default': 'rising', 'values': ('rising', 'falling') }
    )
    annotations = tuple(('ch%d' % i, 'Ch%d' % i) for i in range(MAX_CHANNELS))
    annotation_rows = tuple(('ch%d' % i, 'Ch%d' % i, (i,)) for i in range(MAX_CHANNELS))

    def __init__(self):
        self.reset()

    def reset(self):
        self.samplerate = None
        self.channels = MAX_CHANNELS
        self.channel = 0
        self.bitdepth = 16
        self.bitcount = 0
        self.samplecount = 0
        self.lastsync = 0
        self.lastframe = 0
        self.data = 0
        self.ss_block = None

    def metdatadata(self, key, value):
        if key == srd.SRD_CONF_SAMPLERATE:
            self.samplerate = value

    def start(self):
        self.out_ann = self.register(srd.OUTPUT_ANN)
        self.bitdepth = self.options['bps']
        self.edge = self.options['edge']

    def decode(self):
        while True:
            # Wait for edge of clock (sample on rising/falling edge).
            clock, frame, data = self.wait({0: self.edge[0]})

            self.data = (self.data << 1) | data
            self.bitcount += 1

            if self.ss_block is not None:
                if self.bitcount >= self.bitdepth:
                    self.bitcount = 0
                    self.channel += 1

                    c1 = 'Channel %d' % self.channel
                    c2 = 'C%d' % self.channel
                    c3 = '%d' % self.channel
                    if self.bitdepth <= 8:
                        v = '%02x' % self.data
                    elif self.bitdepth <= 16:
                        v = '%04x' % self.data
                    else:
                        v = '%08x' % self.data

                    if self.channel < self.channels:
                        ch = self.channel
                    else:
                        ch = 0

                    self.put(self.ss_block, self.samplenum, self.out_ann,
                             [ch, ['%s: %s' % (c1, v), '%s: %s' % (c2, v),
                                   '%s: %s' % (c3, v)]])
                    self.data = 0
                    self.ss_block = self.samplenum
                    self.samplecount += 1

            # Check for new frame.
            # Note, frame may be a single clock, or active for the first
            # sample in the frame.
            if frame != self.lastframe and frame == 1:
                self.channel = 0
                self.bitcount = 0
                self.data = 0
                if self.ss_block is None:
                    self.ss_block = 0

            self.lastframe = frame



##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2018 fenugrec <fenugrec@users.sourceforge.net>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
##

import sigrokdecode as srd

class ChannelError(Exception):
    pass

class Decoder(srd.Decoder):
    api_version = 3
    id = 'mcs48'
    name = 'MCS-48'
    longname = 'Intel MCS-48'
    desc = 'Intel MCS-48 external memory access protocol.'
    license = 'gplv2+'
    inputs = ['logic']
    outputs = []
    tags = ['Retro computing']
    channels = (
        {'id': 'ale', 'name': 'ALE', 'desc': 'Address latch enable'},
        {'id': 'psen', 'name': '/PSEN', 'desc': 'Program store enable'},
    ) + tuple({
        'id': 'd%d' % i,
        'name': 'D%d' % i,
        'desc': 'CPU data line %d' % i
        } for i in range(0, 8)
    ) + tuple({
        'id': 'a%d' % i,
        'name': 'A%d' % i,
        'desc': 'CPU address line %d' % i
        } for i in range(8, 12)
    )
    optional_channels = tuple({
        'id': 'a%d' % i,
        'name': 'A%d' % i,
        'desc': 'CPU address line %d' % i
        } for i in range(12, 13)
    )
    annotations = (
        ('romdata', 'Address:Data'),
    )
    binary = (
        ('romdata', 'AAAA:DD'),
    )
    OFF_ALE, OFF_PSEN = 0, 1
    OFF_DATA_BOT, OFF_DATA_TOP = 2, 10
    OFF_ADDR_BOT, OFF_ADDR_TOP = 10, 14
    OFF_BANK_BOT, OFF_BANK_TOP = 14, 15

    def __init__(self):
        self.reset()

    def reset(self):
        self.addr = 0
        self.addr_s = 0
        self.data = 0
        self.data_s = 0

        # Flag to make sure we get an ALE pulse first.
        self.started = 0

    def start(self):
        self.out_ann = self.register(srd.OUTPUT_ANN)
        self.out_bin = self.register(srd.OUTPUT_BINARY)

    def newaddr(self, addr, data):
        # Falling edge on ALE: reconstruct address.
        self.started = 1
        addr = sum([bit << i for i, bit in enumerate(addr)])
        addr <<= len(data)
        addr |= sum([bit << i for i, bit in enumerate(data)])
        self.addr = addr
        self.addr_s = self.samplenum

    def newdata(self, data):
        # Edge on PSEN: get data.
        data = sum([bit << i for i, bit in enumerate(data)])
        self.data = data
        self.data_s = self.samplenum
        if self.started:
            anntext = '{:04X}:{:02X}'.format(self.addr, self.data)
            self.put(self.addr_s, self.data_s, self.out_ann, [0, [anntext]])
            bindata = self.addr.to_bytes(2, byteorder='big')
            bindata += self.data.to_bytes(1, byteorder='big')
            self.put(self.addr_s, self.data_s, self.out_bin, [0, bindata])

    def decode(self):
        # Address bits above A11 are optional, and are considered to be A12+.
        # This logic needs more adjustment when more bank address pins are
        # to get supported. For now, having just A12 is considered sufficient.
        has_bank = self.has_channel(self.OFF_BANK_BOT)
        bank_pin_count = 1 if has_bank else 0
        # Sample address on the falling ALE edge.
        # Save data on falling edge of PSEN.
        while True:
            pins = self.wait([{self.OFF_ALE: 'f'}, {self.OFF_PSEN: 'r'}])
            data = pins[self.OFF_DATA_BOT:self.OFF_DATA_TOP]
            addr = pins[self.OFF_ADDR_BOT:self.OFF_ADDR_TOP]
            bank = pins[self.OFF_BANK_BOT:self.OFF_BANK_TOP]
            if has_bank:
                addr += bank[:bank_pin_count]
            # Handle those conditions (one or more) that matched this time.
            if self.matched[0]:
                self.newaddr(addr, data)
            if self.matched[1]:
                self.newdata(data)



##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2016 Robert Bosch Car Multimedia GmbH
## Authors: Oleksij Rempel
##              <fixed-term.Oleksij.Rempel@de.bosch.com>
##              <linux@rempel-privat.de>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
##

import sigrokdecode as srd

class Decoder(srd.Decoder):
    api_version = 3
    id = 'ssi32'
    name = 'SSI32'
    longname = 'Synchronous Serial Interface (32bit)'
    desc = 'Synchronous Serial Interface (32bit) protocol.'
    license = 'gplv2+'
    inputs = ['spi']
    outputs = []
    tags = ['Embedded/industrial']
    options = (
        {'id': 'msgsize', 'desc': 'Message size', 'default': 64},
    )
    annotations = (
        ('ctrl-tx', 'CTRL TX'),
        ('ack-tx', 'ACK TX'),
        ('ctrl-rx', 'CTRL RX'),
        ('ack-rx', 'ACK RX'),
    )
    annotation_rows = (
        ('tx', 'TX', (0, 1)),
        ('rx', 'RX', (2, 3)),
    )

    def __init__(self):
        self.reset()

    def reset(self):
        self.ss_cmd, self.es_cmd = 0, 0
        self.mosi_bytes = []
        self.miso_bytes = []
        self.es_array = []
        self.rx_size = 0
        self.tx_size = 0

    def start(self):
        self.out_ann = self.register(srd.OUTPUT_ANN)

    def putx(self, data):
        self.put(self.ss_cmd, self.es_cmd, self.out_ann, data)

    def reset_data(self):
        self.mosi_bytes = []
        self.miso_bytes = []
        self.es_array = []

    def handle_ack(self):
        # Only first byte should have ACK data, other 3 bytes are reserved.
        self.es_cmd = self.es_array[0]
        self.putx([1, ['> ACK:0x%02x' % (self.mosi_bytes[0])]])
        self.putx([3, ['< ACK:0x%02x' % (self.miso_bytes[0])]])

    def handle_ctrl(self):
        mosi = miso = ''
        self.tx_size = self.mosi_bytes[2]
        self.rx_size = self.miso_bytes[2]

        if self.tx_size > 0:
            mosi = ', DATA:0x' + ''.join(format(x, '02x') for x in self.mosi_bytes[4:self.tx_size + 4])
        if self.rx_size > 0:
            miso = ', DATA:0x' + ''.join(format(x, '02x') for x in self.miso_bytes[4:self.rx_size + 4])

        self.es_cmd = self.es_array[self.tx_size + 3]
        self.putx([0, ['> CTRL:0x%02x, LUN:0x%02x, SIZE:0x%02x, CRC:0x%02x%s'
                   % (self.mosi_bytes[0], self.mosi_bytes[1],
                      self.mosi_bytes[2], self.mosi_bytes[3], mosi)]])

        self.es_cmd = self.es_array[self.rx_size + 3]
        self.putx([2, ['< CTRL:0x%02x, LUN:0x%02x, SIZE:0x%02x, CRC:0x%02x%s'
                   % (self.miso_bytes[0], self.miso_bytes[1],
                      self.miso_bytes[2], self.miso_bytes[3], miso)]])

    def decode(self, ss, es, data):
        ptype = data[0]
        if ptype == 'CS-CHANGE':
            self.reset_data()
            return

        # Don't care about anything else.
        if ptype != 'DATA':
            return
        mosi, miso = data[1:]

        self.ss, self.es = ss, es

        if len(self.mosi_bytes) == 0:
            self.ss_cmd = ss
        self.mosi_bytes.append(mosi)
        self.miso_bytes.append(miso)
        self.es_array.append(es)

        if self.mosi_bytes[0] & 0x80:
            if len(self.mosi_bytes) < 4:
                return

            self.handle_ack()
            self.reset_data()
        else:
            if len(self.mosi_bytes) < self.options['msgsize']:
                return

            self.handle_ctrl()
            self.reset_data()



##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2019 Marco Geisler <m-sigrok@mageis.de>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
##

regs = {
#   addr: 'name'
    0x00: 'IOCFG2',
    0x01: 'IOCFG1',
    0x02: 'IOCFG0',
    0x03: 'FIFOTHR',
    0x04: 'SYNC1',
    0x05: 'SYNC0',
    0x06: 'PKTLEN',
    0x07: 'PKTCTRL1',
    0x08: 'PKTCTRL0',
    0x09: 'ADDR',
    0x0A: 'CHANNR',
    0x0B: 'FSCTRL1',
    0x0C: 'FSCTRL0',
    0x0D: 'FREQ2',
    0x0E: 'FREQ1',
    0x0F: 'FREQ0',
    0x10: 'MDMCFG4',
    0x11: 'MDMCFG3',
    0x12: 'MDMCFG2',
    0x13: 'MDMCFG1',
    0x14: 'MDMCFG0',
    0x15: 'DEVIATN',
    0x16: 'MCSM2',
    0x17: 'MCSM1',
    0x18: 'MCSM0',
    0x19: 'FOCCFG',
    0x1A: 'BSCFG',
    0x1B: 'AGCTRL2',
    0x1C: 'AGCTRL1',
    0x1D: 'AGCTRL0',
    0x1E: 'WOREVT1',
    0x1F: 'WOREVT0',
    0x20: 'WORCTRL',
    0x21: 'FREND1',
    0x22: 'FREND0',
    0x23: 'FSCAL3',
    0x24: 'FSCAL2',
    0x25: 'FSCAL1',
    0x26: 'FSCAL0',
    0x27: 'RCCTRL1',
    0x28: 'RCCTRL0',
    0x29: 'FSTEST',
    0x2A: 'PTEST',
    0x2B: 'AGCTEST',
    0x2C: 'TEST2',
    0x2D: 'TEST1',
    0x2E: 'TEST0',
    0x30: 'PARTNUM',
    0x31: 'VERSION',
    0x32: 'FREQEST',
    0x33: 'LQI',
    0x34: 'RSSI',
    0x35: 'MARCSTATE',
    0x36: 'WORTIME1',
    0x37: 'WORTIME0',
    0x38: 'PKTSTATUS',
    0x39: 'VCO_VC_DAC',
    0x3A: 'TXBYTES',
    0x3B: 'RXBYTES',
    0x3C: 'RCCTRL1_STATUS',
    0x3D: 'RCCTRL0_STATUS',
    0x3E: 'PATABLE',
    0x3F: 'FIFO'
}

strobes = {
#   addr: 'name'
    0x30: 'SRES',
    0x31: 'SFSTXON',
    0x32: 'SXOFF',
    0x33: 'SCAL',
    0x34: 'SRX',
    0x35: 'STX',
    0x36: 'SIDLE',
    0x37: '',
    0x38: 'SWOR',
    0x39: 'SPWD',
    0x3A: 'SFRX',
    0x3B: 'SFTX',
    0x3C: 'SWORRST',
    0x3D: 'SNOP'
}

status_reg_states = {
#   value: 'state name'
    0b000: 'IDLE',
    0b001: 'RX',
    0b010: 'TX',
    0b011: 'FSTXON',
    0b100: 'CALIBRATE',
    0b101: 'SETTLING',
    0b110: 'RXFIFO_OVERFLOW',
    0b111: 'TXFIFO_OVERFLOW'
}



##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2011 Gareth McMullin <gareth@blacksphere.co.nz>
## Copyright (C) 2012-2014 Uwe Hermann <uwe@hermann-uwe.de>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
##

import sigrokdecode as srd

'''
OUTPUT_PYTHON format:

Packet:
[<ptype>, <pdata>]

<ptype>, <pdata>:
 - 'SYNC', <sync>
 - 'PID', <pid>
 - 'ADDR', <addr>
 - 'EP', <ep>
 - 'CRC5', <crc5>
 - 'CRC5 ERROR', <crc5>
 - 'CRC16', <crc16>
 - 'CRC16 ERROR', <crc16>
 - 'EOP', <eop>
 - 'FRAMENUM', <framenum>
 - 'DATABYTE', <databyte>
 - 'HUBADDR', <hubaddr>
 - 'SC', <sc>
 - 'PORT', <port>
 - 'S', <s>
 - 'E/U', <e/u>
 - 'ET', <et>
 - 'PACKET', [<pcategory>, <pname>, <pinfo>]

<pcategory>, <pname>, <pinfo>:
 - 'TOKEN', 'OUT', [<sync>, <pid>, <addr>, <ep>, <crc5>, <eop>]
 - 'TOKEN', 'IN', [<sync>, <pid>, <addr>, <ep>, <crc5>, <eop>]
 - 'TOKEN', 'SOF', [<sync>, <pid>, <framenum>, <crc5>, <eop>]
 - 'TOKEN', 'SETUP', [<sync>, <pid>, <addr>, <ep>, <crc5>, <eop>]
 - 'DATA', 'DATA0', [<sync>, <pid>, <databytes>, <crc16>, <eop>]
 - 'DATA', 'DATA1', [<sync>, <pid>, <databytes>, <crc16>, <eop>]
 - 'DATA', 'DATA2', [<sync>, <pid>, <databytes>, <crc16>, <eop>]
 - 'DATA', 'MDATA', [<sync>, <pid>, <databytes>, <crc16>, <eop>]
 - 'HANDSHAKE', 'ACK', [<sync>, <pid>, <eop>]
 - 'HANDSHAKE', 'NAK', [<sync>, <pid>, <eop>]
 - 'HANDSHAKE', 'STALL', [<sync>, <pid>, <eop>]
 - 'HANDSHAKE', 'NYET', [<sync>, <pid>, <eop>]
 - 'SPECIAL', 'PRE', [<sync>, <pid>, <addr>, <ep>, <crc5>, <eop>]
 - 'SPECIAL', 'ERR', [<sync>, <pid>, <eop>]
 - 'SPECIAL', 'SPLIT',
   [<sync>, <pid>, <hubaddr>, <sc>, <port>, <s>, <e/u>, <et>, <crc5>, <eop>]
 - 'SPECIAL', 'PING', [<sync>, <pid>, <addr>, <ep>, <crc5>, <eop>]
 - 'SPECIAL', 'Reserved', None

<sync>: SYNC field bitstring, normally '00000001' (8 chars).
<pid>: Packet ID bitstring, e.g. '11000011' for DATA0 (8 chars).
<addr>: Address field number, 0-127 (7 bits).
<ep>: Endpoint number, 0-15 (4 bits).
<crc5>: CRC-5 number (5 bits).
<crc16>: CRC-16 number (16 bits).
<eop>: End of packet marker. List of symbols, usually ['SE0', 'SE0', 'J'].
<framenum>: USB (micro)frame number, 0-2047 (11 bits).
<databyte>: A single data byte, e.g. 0x55.
<databytes>: List of data bytes, e.g. [0x55, 0xaa, 0x99] (0 - 1024 bytes).
<hubaddr>: TODO
<sc>: TODO
<port>: TODO
<s>: TODO
<e/u>: TODO
<et>: TODO
'''

# Packet IDs (PIDs).
# The first 4 bits are the 'packet type' field, the last 4 bits are the
# 'check field' (each bit in the check field must be the inverse of the resp.
# bit in the 'packet type' field; if not, that's a 'PID error').
# For the 4-bit strings, the left-most '1' or '0' is the LSB, i.e. it's sent
# to the bus first.
pids = {
    # Tokens
    '10000111': ['OUT', 'Address & EP number in host-to-function transaction'],
    '10010110': ['IN', 'Address & EP number in function-to-host transaction'],
    '10100101': ['SOF', 'Start-Of-Frame marker & frame number'],
    '10110100': ['SETUP', 'Address & EP number in host-to-function transaction for SETUP to a control pipe'],

    # Data
    # Note: DATA2 and MDATA are HS-only.
    '11000011': ['DATA0', 'Data packet PID even'],
    '11010010': ['DATA1', 'Data packet PID odd'],
    '11100001': ['DATA2', 'Data packet PID HS, high bandwidth isosynchronous transaction in a microframe'],
    '11110000': ['MDATA', 'Data packet PID HS for split and high-bandwidth isosynchronous transactions'],

    # Handshake
    '01001011': ['ACK', 'Receiver accepts error-free packet'],
    '01011010': ['NAK', 'Receiver cannot accept or transmitter cannot send'],
    '01111000': ['STALL', 'EP halted or control pipe request unsupported'],
    '01101001': ['NYET', 'No response yet from receiver'],

    # Special
    '00111100': ['PRE', 'Host-issued preamble; enables downstream bus traffic to low-speed devices'],
    #'00111100': ['ERR', 'Split transaction error handshake'],
    '00011110': ['SPLIT', 'HS split transaction token'],
    '00101101': ['PING', 'HS flow control probe for a bulk/control EP'],
    '00001111': ['Reserved', 'Reserved PID'],
}

def get_category(pidname):
    if pidname in ('OUT', 'IN', 'SOF', 'SETUP'):
        return 'TOKEN'
    elif pidname in ('DATA0', 'DATA1', 'DATA2', 'MDATA'):
        return 'DATA'
    elif pidname in ('ACK', 'NAK', 'STALL', 'NYET'):
        return 'HANDSHAKE'
    else:
        return 'SPECIAL'

def ann_index(pidname):
    l = ['OUT', 'IN', 'SOF', 'SETUP', 'DATA0', 'DATA1', 'DATA2', 'MDATA',
         'ACK', 'NAK', 'STALL', 'NYET', 'PRE', 'ERR', 'SPLIT', 'PING',
         'Reserved']
    if pidname not in l:
        return 28
    return l.index(pidname) + 11

def bitstr_to_num(bitstr):
    if not bitstr:
        return 0
    l = list(bitstr)
    l.reverse()
    return int(''.join(l), 2)

def reverse_number(num, count):
    out = list(count * '0')
    for i in range(0, count):
        if num >> i & 1:
            out[i] = '1'
    return int(''.join(out), 2)

def calc_crc5(bitstr):
    poly5 = 0x25
    crc5 = 0x1f
    for bit in bitstr:
        crc5 <<= 1
        if int(bit) != (crc5 >> 5):
            crc5 ^= poly5
        crc5 &= 0x1f
    crc5 ^= 0x1f
    return reverse_number(crc5, 5)

def calc_crc16(bitstr):
    poly16 = 0x18005
    crc16 = 0xffff
    for bit in bitstr:
        crc16 <<= 1
        if int(bit) != (crc16 >> 16):
            crc16 ^= poly16
        crc16 &= 0xffff
    crc16 ^= 0xffff
    return reverse_number(crc16, 16)

class Decoder(srd.Decoder):
    api_version = 3
    id = 'usb_packet'
    name = 'USB packet'
    longname = 'Universal Serial Bus (LS/FS) packet'
    desc = 'USB (low-speed and full-speed) packet protocol.'
    license = 'gplv2+'
    inputs = ['usb_signalling']
    outputs = ['usb_packet']
    tags = ['PC']
    options = (
        {'id': 'signalling', 'desc': 'Signalling',
            'default': 'full-speed', 'values': ('full-speed', 'low-speed')},
    )
    annotations = (
        ('sync-ok', 'SYNC'),
        ('sync-err', 'SYNC (error)'),
        ('pid', 'PID'),
        ('framenum', 'FRAMENUM'),
        ('addr', 'ADDR'),
        ('ep', 'EP'),
        ('crc5-ok', 'CRC5'),
        ('crc5-err', 'CRC5 (error)'),
        ('data', 'DATA'),
        ('crc16-ok', 'CRC16'),
        ('crc16-err', 'CRC16 (error)'),
        ('packet-out', 'Packet: OUT'),
        ('packet-in', 'Packet: IN'),
        ('packet-sof', 'Packet: SOF'),
        ('packet-setup', 'Packet: SETUP'),
        ('packet-data0', 'Packet: DATA0'),
        ('packet-data1', 'Packet: DATA1'),
        ('packet-data2', 'Packet: DATA2'),
        ('packet-mdata', 'Packet: MDATA'),
        ('packet-ack', 'Packet: ACK'),
        ('packet-nak', 'Packet: NAK'),
        ('packet-stall', 'Packet: STALL'),
        ('packet-nyet', 'Packet: NYET'),
        ('packet-pre', 'Packet: PRE'),
        ('packet-err', 'Packet: ERR'),
        ('packet-split', 'Packet: SPLIT'),
        ('packet-ping', 'Packet: PING'),
        ('packet-reserved', 'Packet: Reserved'),
        ('packet-invalid', 'Packet: Invalid'),
    )
    annotation_rows = (
        ('fields', 'Packet fields', tuple(range(10 + 1))),
        ('packet', 'Packets', tuple(range(11, 28 + 1))),
    )

    def __init__(self):
        self.reset()

    def reset(self):
        self.bits = []
        self.packet = []
        self.packet_summary = ''
        self.ss = self.es = None
        self.ss_packet = self.es_packet = None
        self.state = 'WAIT FOR SOP'

    def putpb(self, data):
        self.put(self.ss, self.es, self.out_python, data)

    def putb(self, data):
        self.put(self.ss, self.es, self.out_ann, data)

    def putpp(self, data):
        self.put(self.ss_packet, self.es_packet, self.out_python, data)

    def putp(self, data):
        self.put(self.ss_packet, self.es_packet, self.out_ann, data)

    def start(self):
        self.out_python = self.register(srd.OUTPUT_PYTHON)
        self.out_ann = self.register(srd.OUTPUT_ANN)

    def handle_packet(self):
        packet = ''
        for (bit, ss, es) in self.bits:
            packet += bit

        if len(packet) < 8:
            self.putp([28, ['Invalid packet (shorter than 8 bits)']])
            return

        # Bits[0:7]: SYNC
        sync = packet[:7 + 1]
        self.ss, self.es = self.bits[0][1], self.bits[7][2]
        # The SYNC pattern for low-speed/full-speed is KJKJKJKK (00000001).
        if sync != '00000001':
            self.putpb(['SYNC ERROR', sync])
            self.putb([1, ['SYNC ERROR: %s' % sync, 'SYNC ERR: %s' % sync,
                           'SYNC ERR', 'SE', 'S']])
        else:
            self.putpb(['SYNC', sync])
            self.putb([0, ['SYNC: %s' % sync, 'SYNC', 'S']])
        self.packet.append(sync)

        if len(packet) < 16:
            self.putp([28, ['Invalid packet (shorter than 16 bits)']])
            return

        # Bits[8:15]: PID
        pid = packet[8:15 + 1]
        pidname = pids.get(pid, ('UNKNOWN', 'Unknown PID'))[0]
        self.ss, self.es = self.bits[8][1], self.bits[15][2]
        self.putpb(['PID', pidname])
        self.putb([2, ['PID: %s' % pidname, pidname, pidname[0]]])
        self.packet.append(pid)
        self.packet_summary += pidname

        if pidname in ('OUT', 'IN', 'SOF', 'SETUP', 'PING'):
            if len(packet) < 32:
                self.putp([28, ['Invalid packet (shorter than 32 bits)']])
                return

            if pidname == 'SOF':
                # Bits[16:26]: Framenum
                framenum = bitstr_to_num(packet[16:26 + 1])
                self.ss, self.es = self.bits[16][1], self.bits[26][2]
                self.putpb(['FRAMENUM', framenum])
                self.putb([3, ['Frame: %d' % framenum, 'Frame', 'Fr', 'F']])
                self.packet.append(framenum)
                self.packet_summary += ' %d' % framenum
            else:
                # Bits[16:22]: Addr
                addr = bitstr_to_num(packet[16:22 + 1])
                self.ss, self.es = self.bits[16][1], self.bits[22][2]
                self.putpb(['ADDR', addr])
                self.putb([4, ['Address: %d' % addr, 'Addr: %d' % addr,
                               'Addr', 'A']])
                self.packet.append(addr)
                self.packet_summary += ' ADDR %d' % addr

                # Bits[23:26]: EP
                ep = bitstr_to_num(packet[23:26 + 1])
                self.ss, self.es = self.bits[23][1], self.bits[26][2]
                self.putpb(['EP', ep])
                self.putb([5, ['Endpoint: %d' % ep, 'EP: %d' % ep, 'EP', 'E']])
                self.packet.append(ep)
                self.packet_summary += ' EP %d' % ep

            # Bits[27:31]: CRC5
            crc5 = bitstr_to_num(packet[27:31 + 1])
            crc5_calc = calc_crc5(packet[16:27])
            self.ss, self.es = self.bits[27][1], self.bits[31][2]
            if crc5 == crc5_calc:
                self.putpb(['CRC5', crc5])
                self.putb([6, ['CRC5: 0x%02X' % crc5, 'CRC5', 'C']])
            else:
                self.putpb(['CRC5 ERROR', crc5])
                self.putb([7, ['CRC5 ERROR: 0x%02X' % crc5, 'CRC5 ERR', 'CE', 'C']])
            self.packet.append(crc5)
        elif pidname in ('DATA0', 'DATA1', 'DATA2', 'MDATA'):
            # Bits[16:packetlen-16]: Data
            data = packet[16:-16]
            # TODO: len(data) must be a multiple of 8.
            databytes = []
            self.packet_summary += ' ['
            for i in range(0, len(data), 8):
                db = bitstr_to_num(data[i:i + 8])
                self.ss, self.es = self.bits[16 + i][1], self.bits[23 + i][2]
                self.putpb(['DATABYTE', db])
                self.putb([8, ['Databyte: %02X' % db, 'Data: %02X' % db,
                               'DB: %02X' % db, '%02X' % db]])
                databytes.append(db)
                self.packet_summary += ' %02X' % db
            self.packet_summary += ' ]'

            # Convenience Python output (no annotation) for all bytes together.
            self.ss, self.es = self.bits[16][1], self.bits[-16][2]
            self.putpb(['DATABYTES', databytes])
            self.packet.append(databytes)

            # Bits[packetlen-16:packetlen]: CRC16
            crc16 = bitstr_to_num(packet[-16:])
            crc16_calc = calc_crc16(packet[16:-16])
            self.ss, self.es = self.bits[-16][1], self.bits[-1][2]
            if crc16 == crc16_calc:
                self.putpb(['CRC16', crc16])
                self.putb([9, ['CRC16: 0x%04X' % crc16, 'CRC16', 'C']])
            else:
                self.putpb(['CRC16 ERROR', crc16])
                self.putb([10, ['CRC16 ERROR: 0x%04X' % crc16, 'CRC16 ERR', 'CE', 'C']])
            self.packet.append(crc16)
        elif pidname in ('ACK', 'NAK', 'STALL', 'NYET', 'ERR'):
            pass # Nothing to do, these only have SYNC+PID+EOP fields.
        elif pidname in ('PRE'):
            pass # Nothing to do, PRE only has SYNC+PID fields.
        else:
            pass # TODO: Handle 'SPLIT' and possibly 'Reserved' packets.

        # Output a (summary of) the whole packet.
        pcategory, pname, pinfo = get_category(pidname), pidname, self.packet
        self.putpp(['PACKET', [pcategory, pname, pinfo]])
        self.putp([ann_index(pidname), ['%s' % self.packet_summary]])

        self.packet, self.packet_summary = [], ''

    def decode(self, ss, es, data):
        (ptype, pdata) = data

        # We only care about certain packet types for now.
        if ptype not in ('SOP', 'BIT', 'EOP', 'ERR'):
            return

        # State machine.
        if self.state == 'WAIT FOR SOP':
            if ptype != 'SOP':
                return
            self.ss_packet = ss
            self.state = 'GET BIT'
        elif self.state == 'GET BIT':
            if ptype == 'BIT':
                self.bits.append([pdata, ss, es])
            elif ptype == 'EOP' or ptype == 'ERR':
                self.es_packet = es
                self.handle_packet()
                self.packet, self.packet_summary = [], ''
                self.bits, self.state = [], 'WAIT FOR SOP'
            else:
                pass # TODO: Error



##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2012 Uwe Hermann <uwe@hermann-uwe.de>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
##

'''
This decoder stacks on top of the 'spi' PD and decodes the SD card
(SPI mode) low-level protocol.

Most SD cards can be accessed via two different protocols/modes: SD mode
or SPI mode.

All SD cards are in SD mode upon powerup. They can be switched to SPI mode
using a special method involving CMD0 (see spec). Once in SPI mode, the mode
can no longer be changed without a power-cycle of the card.

SPI mode properties (differences to SD mode):
 * The 'sdcard_spi' PD stacks on top of the 'spi' PD. This is not possible
   for the 'sdcard_sd' PD, as that protocol is not SPI related at all.
   Hence 'sdcard_spi' and 'sdcard_sd' are two separate PDs.
 * The state machines for SPI mode and SD mode are different.
 * In SPI mode, data transfers are byte-oriented (commands/data are multiples
   of 8 bits, with the CS# pin asserted respectively), unlike SD mode where
   commands/data are bit-oriented with parallel transmission of 1 or 4 bits.
 * While the SPI mode command set has some commands in common with the
   SD mode command set, they are not the same and also not a subset/superset.
   Some commands are only available in SD mode (e.g. CMD2), some only
   in SPI mode (e.g. CMD1).
 * Response types of commands also differ between SD mode and SPI mode.
   E.g. CMD9 has an R2 response in SD mode, but R1 in SPI mode.
 * The commands and functions in SD mode defined after version 2.00 of the
   spec are NOT supported in SPI mode.
 * SPI mode: The selected SD card ALWAYS responds to commands (unlike SD mode).
 * Upon data retrieval problems (read operations) the card will respond with
   an error response (and no data), as opposed to a timeout in SD mode.
 * SPI mode: For every data block sent to the card (write operations) the host
   gets a data response token from the card.
 * SDSC: A data block can be max. one card write block, min. 1 byte.
 * SDHC/SDXC: Block length is fixed to 512. The block length set by CMD16
   is only used for CDM42 (not for memory data transfers). Thus, partial
   read/write operations are disabled in SPI mode.
 * SPI mode: Write protected commands (CMD28, CMD29, CMD30) are not supported.
 * The SD mode state machine is NOT used. All commands that are supported
   in SPI mode are always available.
 * Per default the card is in CRC OFF mode. Exception: CMD0 (which is used to
   switch to SPI mode) needs a valid CRC.
 * The APP_CMD status bit is not available in SPI mode.
 * TODO: Switch function command differences.
 * In SPI mode cards cannot guarantee their speed class (the host should
   assume class 0, no matter what the card indicates).
 * The RCA register is not accessible in SPI mode.
'''

from .pd import Decoder



##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2013 Bert Vermeulen <bert@biot.com>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
##

'''
This PD decodes the XFP I²C management interface structures/protocol.

XFP modules include an I²C interface, used to monitor and control various
aspects of the module. The specification defines an I²C slave at address
0x50 (0xa0) which returns 128 bytes of a standard structure ("lower memory"),
and, after setting a table number in lower memory, a set of 256 "higher
memory" tables, which can be mapped to different subdevices on the XFP.

Only one table is defined in the specification: table 0x01, the default on
module startup. Other table are either reserved for future expansion, or
available for vendor-specific extensions. This decoder supports both lower
memory and table 0x01.

Details:
ftp://ftp.seagate.com/sff/INF-8077.PDF (XFP specification)
'''

from .pd import Decoder



##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2011-2014 Uwe Hermann <uwe@hermann-uwe.de>
## Copyright (C) 2016 Gerhard Sittig <gerhard.sittig@gmx.net>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
##

# Note the implementation details:
#
# Although the Atmel literature suggests (does not explicitly mandate,
# but shows in diagrams) that two stop bits are used in the protocol,
# the decoder loses synchronization with ATxmega generated responses
# when it expects more than one stop bit. Since the chip's hardware is
# fixed, this is not an implementation error in some programmer software.
# Since this is a protocol decoder which does not participate in the
# communication (does not actively send data), we can read the data
# stream with one stop bit, and transparently keep working when two
# are used.
#
# Annotations in the UART fields level differ from Atmel literature.
# Wrong parity bits are referred to as "parity error". Low stop bits are
# referred to as "frame error".
#
# The PDI component in the device starts disabled. Enabling PDI
# communication is done by raising DATA and clocking RESET with a
# minimum frequency. PDI communication automatically gets disabled when
# RESET "is inactive" for a certain period of time. The specific timing
# conditions are rather fuzzy in the literature (phrased weakly), and
# are device dependent (refer to the minumum RESET pulse width). This
# protocol decoder implementation internally prepares for but currently
# does not support these enable and disable phases. On the one hand it
# avoids excess external dependencies or wrong results for legal input
# data. On the other hand the decoder works when input streams start in
# the middle of an established connection.
#
# Communication peers detect physical collisions. The decoder can't.
# Upon collisions, a peer will cease any subsequent transmission, until
# a BREAK is seen. Synchronization can get enforced by sending two BREAK
# conditions. The first will cause a collision, the second will re-enable
# the peer. The decoder has no concept of physical collisions. It stops
# the interpretation of instructions when BREAK is seen, and assumes
# that a new instruction will start after BREAK.
#
# This protocol decoder only supports PDI communication over UART frames.
# It lacks support for PDI over JTAG. This would require separation into
# multiple protocol decoder layers (UART physical, JTAG physical, PDI
# instructions, optionally device support on top of PDI. There is some
# more potential for future extensions:
# - The JTAG physical has dedicated TX and RX directions. This decoder
#   only picks up communicated bytes but does not check which "line"
#   they are communicated on (not applicable to half duplex UART).
# - PDI over JTAG uses "special frame error" conditions to communicate
#   additional symbols: BREAK (0xBB with parity 1), DELAY (0xDB with
#   parity 1), and EMPTY (0xEB with parity 1).
# - Another "device support" layer might interpret device specific
#   timings, and might map addresses used in memory access operations
#   to component names, or even register names and bit fields(?). It's
#   quite deep a rabbithole though...

import sigrokdecode as srd
from collections import namedtuple

class Ann:
    '''Annotation and binary output classes.'''
    (
        BIT, START, DATA, PARITY_OK, PARITY_ERR,
        STOP_OK, STOP_ERR, BREAK,
        OPCODE, DATA_PROG, DATA_DEV, PDI_BREAK,
        ENABLE, DISABLE, COMMAND,
    ) = range(15)
    (
        BIN_BYTES,
    ) = range(1)

Bit = namedtuple('Bit', 'val ss es')

class PDI:
    '''PDI protocol instruction opcodes, and operand formats.'''
    (
        OP_LDS, OP_LD, OP_STS, OP_ST,
        OP_LDCS, OP_REPEAT, OP_STCS, OP_KEY,
    ) = range(8)
    pointer_format_nice = [
        '*(ptr)',
        '*(ptr++)',
        'ptr',
        'ptr++ (rsv)',
    ]
    pointer_format_terse = [
        '*p',
        '*p++',
        'p',
        '(rsv)',
    ]
    ctrl_reg_name = {
        0: 'status',
        1: 'reset',
        2: 'ctrl',
    }

class Decoder(srd.Decoder):
    api_version = 3
    id = 'avr_pdi'
    name = 'AVR PDI'
    longname = 'Atmel Program and Debug Interface'
    desc = 'Atmel ATxmega Program and Debug Interface (PDI) protocol.'
    license = 'gplv2+'
    inputs = ['logic']
    outputs = []
    tags = ['Debug/trace']
    channels = (
        {'id': 'reset', 'name': 'RESET', 'desc': 'RESET / PDI_CLK'},
        {'id': 'data', 'name': 'DATA', 'desc': 'PDI_DATA'},
    )
    annotations = (
        ('uart-bit', 'UART bit'),
        ('start-bit', 'Start bit'),
        ('data-bit', 'Data bit'),
        ('parity-ok', 'Parity OK bit'),
        ('parity-err', 'Parity error bit'),
        ('stop-ok', 'Stop OK bit'),
        ('stop-err', 'Stop error bit'),
        ('break', 'BREAK condition'),
        ('opcode', 'Instruction opcode'),
        ('data-prog', 'Programmer data'),
        ('data-dev', 'Device data'),
        ('pdi-break', 'BREAK at PDI level'),
        ('enable', 'Enable PDI'),
        ('disable', 'Disable PDI'),
        ('cmd-data', 'PDI command with data'),
    )
    annotation_rows = (
        ('uart_bits', 'UART bits', (Ann.BIT,)),
        ('uart_fields', 'UART fields', (Ann.START, Ann.DATA, Ann.PARITY_OK,
            Ann.PARITY_ERR, Ann.STOP_OK, Ann.STOP_ERR, Ann.BREAK)),
        ('pdi_fields', 'PDI fields', (Ann.OPCODE, Ann.DATA_PROG, Ann.DATA_DEV,
            Ann.PDI_BREAK)),
        ('pdi_cmds', 'PDI Cmds', (Ann.ENABLE, Ann.DISABLE, Ann.COMMAND)),
    )
    binary = (
        ('bytes', 'PDI protocol bytes'),
    )

    def __init__(self):
        self.reset()

    def reset(self):
        self.samplerate = None
        self.clear_state()

    def clear_state(self):
        # Track bit times and bit values.
        self.ss_last_fall = None
        self.data_sample = None
        self.ss_curr_fall = None
        # Collect UART frame bits into byte values.
        self.bits = []
        self.zero_count = 0
        self.zero_ss = None
        self.break_ss = None
        self.break_es = None
        self.clear_insn()

    def clear_insn(self):
        # Collect instructions and their arguments,
        # properties of the current instructions.
        self.insn_rep_count = 0
        self.insn_opcode = None
        self.insn_wr_counts = []
        self.insn_rd_counts = []
        # Accumulation of data items as bytes pass by.
        self.insn_dat_bytes = []
        self.insn_dat_count = 0
        self.insn_ss_data = None
        # Next layer "commands", instructions plus operands.
        self.cmd_ss = None
        self.cmd_insn_parts_nice = []
        self.cmd_insn_parts_terse = []

    def metadata(self, key, value):
        if key == srd.SRD_CONF_SAMPLERATE:
            self.samplerate = value

    def start(self):
        self.out_ann = self.register(srd.OUTPUT_ANN)
        self.out_binary = self.register(srd.OUTPUT_BINARY)

    def put_ann_bit(self, bit_nr, ann_idx):
        b = self.bits[bit_nr]
        self.put(b.ss, b.es, self.out_ann, [ann_idx, [str(b.val)]])

    def put_ann_data(self, bit_nr, ann_data):
        b = self.bits[bit_nr]
        self.put(b.ss, b.es, self.out_ann, ann_data)

    def put_ann_row_val(self, ss, es, row, value):
        self.put(ss, es, self.out_ann, [row, value])

    def put_bin_bytes(self, ss, es, row, value):
        self.put(ss, es, self.out_binary, [row, value])

    def handle_byte(self, ss, es, byteval):
        '''Handle a byte at the PDI protocol layer.'''

        # Handle BREAK conditions, which will abort any
        # potentially currently executing instruction.
        is_break = byteval is None
        if is_break:
            self.cmd_insn_parts_nice.append('BREAK')
            self.cmd_insn_parts_terse.append('BRK')
            self.insn_rep_count = 0
            # Will FALLTHROUGH to "end of instruction" below.

        # Decode instruction opcodes and argument sizes
        # from the first byte of a transaction.
        if self.insn_opcode is None and not is_break:
            opcode = (byteval & 0xe0) >> 5
            arg30 = byteval & 0x0f
            arg32 = (byteval & 0x0c) >> 2
            arg10 = byteval & 0x03
            self.insn_opcode = opcode
            self.cmd_ss = ss
            mnemonics = None
            if opcode == PDI.OP_LDS:
                # LDS: load data, direct addressing.
                # Writes an address, reads a data item.
                width_addr = arg32 + 1
                width_data = arg10 + 1
                self.insn_wr_counts = [width_addr]
                self.insn_rd_counts = [width_data]
                mnemonics = [
                    'Insn: LDS a{:d}, m{:d}'.format(width_addr, width_data),
                    'LDS a{:d}, m{:d}'.format(width_addr, width_data), 'LDS',
                ]
                self.cmd_insn_parts_nice = ['LDS']
                self.cmd_insn_parts_terse = ['LDS']
            elif opcode == PDI.OP_LD:
                # LD: load data, indirect addressing.
                # Reads a data item, with optional repeat.
                ptr_txt = PDI.pointer_format_nice[arg32]
                ptr_txt_terse = PDI.pointer_format_terse[arg32]
                width_data = arg10 + 1
                self.insn_wr_counts = []
                self.insn_rd_counts = [width_data]
                if self.insn_rep_count:
                    self.insn_rd_counts.extend(self.insn_rep_count * [width_data])
                    self.insn_rep_count = 0
                mnemonics = [
                    'Insn: LD {:s} m{:d}'.format(ptr_txt, width_data),
                    'LD {:s} m{:d}'.format(ptr_txt, width_data), 'LD',
                ]
                self.cmd_insn_parts_nice = ['LD', ptr_txt]
                self.cmd_insn_parts_terse = ['LD', ptr_txt_terse]
            elif opcode == PDI.OP_STS:
                # STS: store data, direct addressing.
                # Writes an address, writes a data item.
                width_addr = arg32 + 1
                width_data = arg10 + 1
                self.insn_wr_counts = [width_addr, width_data]
                self.insn_rd_counts = []
                mnemonics = [
                    'Insn: STS a{:d}, i{:d}'.format(width_addr, width_data),
                    'STS a{:d}, i{:d}'.format(width_addr, width_data), 'STS',
                ]
                self.cmd_insn_parts_nice = ['STS']
                self.cmd_insn_parts_terse = ['STS']
            elif opcode == PDI.OP_ST:
                # ST: store data, indirect addressing.
                # Writes a data item, with optional repeat.
                ptr_txt = PDI.pointer_format_nice[arg32]
                ptr_txt_terse = PDI.pointer_format_terse[arg32]
                width_data = arg10 + 1
                self.insn_wr_counts = [width_data]
                self.insn_rd_counts = []
                if self.insn_rep_count:
                    self.insn_wr_counts.extend(self.insn_rep_count * [width_data])
                    self.insn_rep_count = 0
                mnemonics = [
                    'Insn: ST {:s} i{:d}'.format(ptr_txt, width_data),
                    'ST {:s} i{:d}'.format(ptr_txt, width_data), 'ST',
                ]
                self.cmd_insn_parts_nice = ['ST', ptr_txt]
                self.cmd_insn_parts_terse = ['ST', ptr_txt_terse]
            elif opcode == PDI.OP_LDCS:
                # LDCS: load control/status.
                # Loads exactly one byte.
                reg_num = arg30
                reg_txt = PDI.ctrl_reg_name.get(reg_num, 'r{:d}'.format(reg_num))
                reg_txt_terse = '{:d}'.format(reg_num)
                self.insn_wr_counts = []
                self.insn_rd_counts = [1]
                mnemonics = [
                    'Insn: LDCS {:s}, m1'.format(reg_txt),
                    'LDCS {:s}, m1'.format(reg_txt), 'LDCS',
                ]
                self.cmd_insn_parts_nice = ['LDCS', reg_txt]
                self.cmd_insn_parts_terse = ['LDCS', reg_txt_terse]
            elif opcode == PDI.OP_STCS:
                # STCS: store control/status.
                # Writes exactly one byte.
                reg_num = arg30
                reg_txt = PDI.ctrl_reg_name.get(reg_num, 'r{:d}'.format(reg_num))
                reg_txt_terse = '{:d}'.format(reg_num)
                self.insn_wr_counts = [1]
                self.insn_rd_counts = []
                mnemonics = [
                    'Insn: STCS {:s}, i1'.format(reg_txt),
                    'STCS {:s}, i1'.format(reg_txt), 'STCS',
                ]
                self.cmd_insn_parts_nice = ['STCS', reg_txt]
                self.cmd_insn_parts_terse = ['STCS', reg_txt_terse]
            elif opcode == PDI.OP_REPEAT:
                # REPEAT: sets repeat count for the next instruction.
                # Reads repeat count from following bytes.
                width_data = arg10 + 1
                self.insn_wr_counts = [width_data]
                self.insn_rd_counts = []
                mnemonics = [
                    'Insn: REPEAT i{:d}'.format(width_data),
                    'REPEAT i{:d}'.format(width_data), 'REP',
                ]
                self.cmd_insn_parts_nice = ['REPEAT']
                self.cmd_insn_parts_terse = ['REP']
            elif opcode == PDI.OP_KEY:
                # KEY: set activation key (enables PDIBUS mmap access).
                # Writes a sequence of 8 bytes, fixed length.
                width_data = 8
                self.insn_wr_counts = [width_data]
                self.insn_rd_counts = []
                mnemonics = [
                    'Insn: KEY i{:d}'.format(width_data),
                    'KEY i{:d}'.format(width_data), 'KEY',
                ]
                self.cmd_insn_parts_nice = ['KEY']
                self.cmd_insn_parts_terse = ['KEY']

            # Emit an annotation for the instruction opcode.
            self.put_ann_row_val(ss, es, Ann.OPCODE, mnemonics)

            # Prepare to write/read operands/data bytes.
            self.insn_dat_bytes = []
            if self.insn_wr_counts:
                self.insn_dat_count = self.insn_wr_counts[0]
                return
            if self.insn_rd_counts:
                self.insn_dat_count = self.insn_rd_counts[0]
                return
            # FALLTHROUGH.
            # When there are no operands or data bytes to read,
            # then fall through to the end of the instruction
            # handling below (which emits annotations).

        # Read bytes which carry operands (addresses, immediates)
        # or data values for memory access.
        if self.insn_dat_count and not is_break:

            # Accumulate received bytes until another multi byte
            # data item is complete.
            if not self.insn_dat_bytes:
                self.insn_ss_data = ss
            self.insn_dat_bytes.append(byteval)
            self.insn_dat_count -= 1
            if self.insn_dat_count:
                return

            # Determine the data item's duration and direction,
            # "consume" its length spec (to simplify later steps).
            data_ss = self.insn_ss_data
            data_es = es
            if self.insn_wr_counts:
                data_ann = Ann.DATA_PROG
                data_width = self.insn_wr_counts.pop(0)
            elif self.insn_rd_counts:
                data_ann = Ann.DATA_DEV
                data_width = self.insn_rd_counts.pop(0)

            # PDI communicates multi-byte data items in little endian
            # order. Get a nice textual representation of the number,
            # wide and narrow for several zoom levels.
            self.insn_dat_bytes.reverse()
            data_txt_digits = ''.join(['{:02x}'.format(b) for b in self.insn_dat_bytes])
            data_txt_hex = '0x' + data_txt_digits
            data_txt_prefix = 'Data: ' + data_txt_hex
            data_txts = [data_txt_prefix, data_txt_hex, data_txt_digits]
            self.insn_dat_bytes = []

            # Emit an annotation for the data value.
            self.put_ann_row_val(data_ss, data_es, data_ann, data_txts)

            # Collect detailled information which describes the whole
            # command when combined (for a next layer annotation,
            # spanning the complete command).
            self.cmd_insn_parts_nice.append(data_txt_hex)
            self.cmd_insn_parts_terse.append(data_txt_digits)

            # Send out write data first until exhausted,
            # then drain expected read data.
            if self.insn_wr_counts:
                self.insn_dat_count = self.insn_wr_counts[0]
                return
            if self.insn_rd_counts:
                self.insn_dat_count = self.insn_rd_counts[0]
                return

            # FALLTHROUGH.
            # When all operands and data bytes were seen,
            # terminate the inspection of the instruction.

        # Postprocess the instruction after its operands were seen.
        cmd_es = es
        cmd_txt_nice = ' '.join(self.cmd_insn_parts_nice)
        cmd_txt_terse = ' '.join(self.cmd_insn_parts_terse)
        cmd_txts = [cmd_txt_nice, cmd_txt_terse]
        self.put_ann_row_val(self.cmd_ss, cmd_es, Ann.COMMAND, cmd_txts)
        if self.insn_opcode == PDI.OP_REPEAT and not is_break:
            # The last communicated data item is the repeat
            # count for the next instruction (i.e. it will
            # execute N+1 times when "REPEAT N" is specified).
            count = int(self.cmd_insn_parts_nice[-1], 0)
            self.insn_rep_count = count

        # Have the state for instruction decoding cleared, but make sure
        # to carry over REPEAT count specs between instructions. They
        # start out as zero, will be setup by REPEAT instructions, need
        # to get passed to the instruction which follows REPEAT. The
        # instruction which sees a non-zero repeat count which will
        # consume the counter and drop it to zero, then the counter
        # remains at zero until the next REPEAT instruction.
        save_rep_count = self.insn_rep_count
        self.clear_insn()
        self.insn_rep_count = save_rep_count

    def handle_bits(self, ss, es, bitval):
        '''Handle a bit at the UART layer.'''

        # Concentrate annotation literals here for easier maintenance.
        ann_class_text = {
            Ann.START: ['Start bit', 'Start', 'S'],
            Ann.PARITY_OK: ['Parity OK', 'Par OK', 'P'],
            Ann.PARITY_ERR: ['Parity error', 'Par ERR', 'PE'],
            Ann.STOP_OK: ['Stop bit', 'Stop', 'T'],
            Ann.STOP_ERR: ['Stop bit error', 'Stop ERR', 'TE'],
            Ann.BREAK: ['Break condition', 'BREAK', 'BRK'],
        }
        def put_uart_field(bitpos, annclass):
            self.put_ann_data(bitpos, [annclass, ann_class_text[annclass]])

        # The number of bits which form one UART frame. Note that
        # the decoder operates with only one stop bit.
        frame_bitcount = 1 + 8 + 1 + 1

        # Detect adjacent runs of all-zero bits. This is meant
        # to cope when BREAK conditions appear at any arbitrary
        # position, it need not be "aligned" to an UART frame.
        if bitval == 1:
            self.zero_count = 0
        elif bitval == 0:
            if not self.zero_count:
                self.zero_ss = ss
            self.zero_count += 1
            if self.zero_count == frame_bitcount:
                self.break_ss = self.zero_ss

        # BREAK conditions are _at_minimum_ the length of a UART frame, but
        # can span an arbitrary number of bit times. Track the "end sample"
        # value of the last low bit we have seen, and emit the annotation only
        # after the line went idle (high) again. Pass BREAK to the upper layer
        # as well. When the line is low, BREAK still is pending. When the line
        # is high, the current bit cannot be START, thus return from here.
        if self.break_ss is not None:
            if bitval == '0':
                self.break_es = es
                return
            self.put(self.break_ss, self.break_es, self.out_ann,
                 [Ann.BREAK, ann_class_text[Ann.BREAK]])
            self.handle_byte(self.break_ss, self.break_es, None)
            self.break_ss = None
            self.break_es = None
            self.bits = []
            return

        # Ignore high bits when waiting for START.
        if not self.bits and bitval == 1:
            return

        # Store individual bits and their start/end sample numbers,
        # until a complete frame was received.
        self.bits.append(Bit(bitval, ss, es))
        if len(self.bits) < frame_bitcount:
            return

        # Get individual fields of the UART frame.
        bits_num = sum([b.val << pos for pos, b in enumerate(self.bits)])
        if False:
            # This logic could detect BREAK conditions which are aligned to
            # UART frames. Which was obsoleted by the above detection at
            # arbitrary positions. The code still can be useful to detect
            # "other kinds of frame errors" which carry valid symbols for
            # upper layers (the Atmel literature suggests "break", "delay",
            # and "empty" symbols when PDI is communicated over different
            # physical layers).
            if bits_num == 0: # BREAK
                self.break_ss = self.bits[0].ss
                self.break_es = es
                self.bits = []
                return
        start_bit = bits_num & 0x01; bits_num >>= 1
        data_val = bits_num & 0xff; bits_num >>= 8
        data_text = '{:02x}'.format(data_val)
        parity_bit = bits_num & 0x01; bits_num >>= 1
        stop_bit = bits_num & 0x01; bits_num >>= 1

        # Check for frame errors. START _must_ have been low
        # according to the above accumulation logic.
        parity_ok = (bin(data_val).count('1') + parity_bit) % 2 == 0
        stop_ok = stop_bit == 1
        valid_frame = parity_ok and stop_ok

        # Emit annotations.
        for idx in range(frame_bitcount):
            self.put_ann_bit(idx, Ann.BIT)
        put_uart_field(0, Ann.START)
        self.put(self.bits[1].ss, self.bits[8].es, self.out_ann,
             [Ann.DATA, ['Data: ' + data_text, 'D: ' + data_text, data_text]])
        put_uart_field(9, Ann.PARITY_OK if parity_ok else Ann.PARITY_ERR)
        put_uart_field(10, Ann.STOP_OK if stop_ok else Ann.STOP_ERR)

        # Emit binary data stream. Have bytes interpreted at higher layers.
        if valid_frame:
            byte_ss, byte_es = self.bits[0].ss, self.bits[-1].es
            self.put_bin_bytes(byte_ss, byte_es, Ann.BIN_BYTES, bytes([data_val]))
            self.handle_byte(byte_ss, byte_es, data_val)

        # Reset internal state for the next frame.
        self.bits = []

    def handle_clk_edge(self, clock_pin, data_pin):
        # Sample the data line on rising clock edges. Always, for TX and for
        # RX bytes alike.
        if clock_pin == 1:
            self.data_sample = data_pin
            return

        # Falling clock edges are boundaries for bit slots. Inspect previously
        # sampled bits on falling clock edges, when the start and end sample
        # numbers were determined. Only inspect bit slots of known clock
        # periods (avoid interpreting the DATA line when the "enabled" state
        # has not yet been determined).
        self.ss_last_fall = self.ss_curr_fall
        self.ss_curr_fall = self.samplenum
        if self.ss_last_fall is None:
            return

        # Have the past bit slot processed.
        bit_ss, bit_es = self.ss_last_fall, self.ss_curr_fall
        bit_val = self.data_sample
        self.handle_bits(bit_ss, bit_es, bit_val)

    def decode(self):
        while True:
            self.handle_clk_edge(*self.wait({0: 'e'}))



##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2016 Bert Vermeulen <bert@biot.com>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
##

# This module contains definitions for use by pluggable network adapters,
# such as SFP, XFP etc.

MODULE_ID = {
    0x01: 'GBIC',
    0x02: 'Integrated module/connector',
    0x03: 'SFP',
    0x04: '300-pin XBI',
    0x05: 'XENPAK',
    0x06: 'XFP',
    0x07: 'XFF',
    0x08: 'XFP-E',
    0x09: 'XPAK',
    0x0a: 'X2',
}

ALARM_THRESHOLDS = {
    0:  'Temp high alarm',
    2:  'Temp low alarm',
    4:  'Temp high warning',
    6:  'Temp low warning',
    16: 'Bias high alarm',
    18: 'Bias low alarm',
    20: 'Bias high warning',
    22: 'Bias low warning',
    24: 'TX power high alarm',
    26: 'TX power low alarm',
    28: 'TX power high warning',
    30: 'TX power low warning',
    32: 'RX power high alarm',
    34: 'RX power low alarm',
    36: 'RX power high warning',
    38: 'RX power low warning',
    40: 'AUX 1 high alarm',
    42: 'AUX 1 low alarm',
    44: 'AUX 1 high warning',
    46: 'AUX 1 low warning',
    48: 'AUX 2 high alarm',
    50: 'AUX 2 low alarm',
    52: 'AUX 2 high warning',
    54: 'AUX 2 low warning',
}

AD_READOUTS = {
    0:  'Module temperature',
    4:  'TX bias current',
    6:  'Measured TX output power',
    8:  'Measured RX input power',
    10: 'AUX 1 measurement',
    12: 'AUX 2 measurement',
}

GCS_BITS = [
    'TX disable',
    'Soft TX disable',
    'MOD_NR',
    'P_Down',
    'Soft P_Down',
    'Interrupt',
    'RX_LOS',
    'Data_Not_Ready',
    'TX_NR',
    'TX_Fault',
    'TX_CDR not locked',
    'RX_NR',
    'RX_CDR not locked',
]

CONNECTOR = {
    0x01:   'SC',
    0x02:   'Fibre Channel style 1 copper',
    0x03:   'Fibre Channel style 2 copper',
    0x04:   'BNC/TNC',
    0x05:   'Fibre Channel coax',
    0x06:   'FiberJack',
    0x07:   'LC',
    0x08:   'MT-RJ',
    0x09:   'MU',
    0x0a:   'SG',
    0x0b:   'Optical pigtail',
    0x20:   'HSSDC II',
    0x21:   'Copper pigtail',
}

TRANSCEIVER = [
    # 10GB Ethernet
    ['10GBASE-SR', '10GBASE-LR', '10GBASE-ER', '10GBASE-LRM', '10GBASE-SW',
        '10GBASE-LW',   '10GBASE-EW'],
    # 10GB Fibre Channel
    ['1200-MX-SN-I', '1200-SM-LL-L', 'Extended Reach 1550 nm',
        'Intermediate reach 1300 nm FP'],
    # 10GB Copper
    [],
    # 10GB low speed
    ['1000BASE-SX / 1xFC MMF', '1000BASE-LX / 1xFC SMF', '2xFC MMF',
        '2xFC SMF', 'OC48-SR', 'OC48-IR', 'OC48-LR'],
    # 10GB SONET/SDH interconnect
    ['I-64.1r', 'I-64.1', 'I-64.2r', 'I-64.2', 'I-64.3', 'I-64.5'],
    # 10GB SONET/SDH short haul
    ['S-64.1', 'S-64.2a', 'S-64.2b', 'S-64.3a', 'S-64.3b', 'S-64.5a', 'S-64.5b'],
    # 10GB SONET/SDH long haul
    ['L-64.1', 'L-64.2a', 'L-64.2b', 'L-64.2c', 'L-64.3', 'G.959.1 P1L1-2D2'],
    # 10GB SONET/SDH very long haul
    ['V-64.2a', 'V-64.2b', 'V-64.3'],
]

SERIAL_ENCODING = [
    '64B/66B',
    '8B/10B',
    'SONET scrambled',
    'NRZ',
    'RZ',
]

XMIT_TECH = [
    '850 nm VCSEL',
    '1310 nm VCSEL',
    '1550 nm VCSEL',
    '1310 nm FP',
    '1310 nm DFB',
    '1550 nm DFB',
    '1310 nm EML'
    '1550 nm EML'
    'copper',
]

CDR = [
    '9.95Gb/s',
    '10.3Gb/s',
    '10.5Gb/s',
    '10.7Gb/s',
    '11.1Gb/s',
    '(unknown)',
    'lineside loopback mode',
    'XFI loopback mode',
]

DEVICE_TECH = [
    ['no wavelength control', 'sctive wavelength control'],
    ['uncooled transmitter device', 'cooled transmitter'],
    ['PIN detector', 'APD detector'],
    ['transmitter not tunable', 'transmitter tunable'],
]

ENHANCED_OPTS = [
    'VPS',
    'soft TX_DISABLE',
    'soft P_Down',
    'VPS LV regulator mode',
    'VPS bypassed regulator mode',
    'active FEC control',
    'wavelength tunability',
    'CMU',
]

AUX_TYPES = [
    'not implemented',
    'APD bias voltage',
    '(unknown)',
    'TEC current',
    'laser temperature',
    'laser wavelength',
    '5V supply voltage',
    '3.3V supply voltage',
    '1.8V supply voltage',
    '-5.2V supply voltage',
    '5V supply current',
    '(unknown)',
    '(unknown)',
    '3.3V supply current',
    '1.8V supply current',
    '-5.2V supply current',
]



##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2017 Kevin Redon <kingkevin@cuvoodoo.info>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
##

import sigrokdecode as srd
from collections import namedtuple

'''
OUTPUT_PYTHON format:

Packet:
[namedtuple('ss': bit start sample number,
  'es': bit end sample number,
  'si': SI bit,
  'so': SO bit,
 ), ...]

Since address and word size are variable, a list of all bits in each packet
need to be output. Since Microwire is a synchronous protocol with separate
input and output lines (SI and SO) they are provided together, but because
Microwire is half-duplex only the SI or SO bits will be considered at once.
To be able to annotate correctly the instructions formed by the bit, the start
and end sample number of each bit (pair of SI/SO bit) are provided.
'''

PyPacket = namedtuple('PyPacket', 'ss es si so')
Packet = namedtuple('Packet', 'samplenum matched cs sk si so')

class Decoder(srd.Decoder):
    api_version = 3
    id = 'microwire'
    name = 'Microwire'
    longname = 'Microwire'
    desc = '3-wire, half-duplex, synchronous serial bus.'
    license = 'gplv2+'
    inputs = ['logic']
    outputs = ['microwire']
    tags = ['Embedded/industrial']
    channels = (
        {'id': 'cs', 'name': 'CS', 'desc': 'Chip select'},
        {'id': 'sk', 'name': 'SK', 'desc': 'Clock'},
        {'id': 'si', 'name': 'SI', 'desc': 'Slave in'},
        {'id': 'so', 'name': 'SO', 'desc': 'Slave out'},
    )
    annotations = (
        ('start-bit', 'Start bit'),
        ('si-bit', 'SI bit'),
        ('so-bit', 'SO bit'),
        ('status-check-ready', 'Status check ready'),
        ('status-check-busy', 'Status check busy'),
        ('warning', 'Warning'),
    )
    annotation_rows = (
        ('si-bits', 'SI bits', (0, 1)),
        ('so-bits', 'SO bits', (2,)),
        ('status', 'Status', (3, 4)),
        ('warnings', 'Warnings', (5,)),
    )

    def __init__(self):
        self.reset()

    def reset(self):
        pass

    def start(self):
        self.out_python = self.register(srd.OUTPUT_PYTHON)
        self.out_ann = self.register(srd.OUTPUT_ANN)

    def decode(self):
        while True:
            # Wait for slave to be selected on rising CS.
            cs, sk, si, so = self.wait({0: 'r'})
            if sk:
                self.put(self.samplenum, self.samplenum, self.out_ann,
                     [5, ['Clock should be low on start',
                     'Clock high on start', 'Clock high', 'SK high']])
                sk = 0 # Enforce correct state for correct clock handling.
                # Because we don't know if this is bit communication or a
                # status check we have to collect the SI and SO values on SK
                # edges while the chip is selected and figure out afterwards.
            packet = []
            while cs:
                # Save change.
                packet.append(Packet(self.samplenum, self.matched, cs, sk, si, so))
                edge = 'r' if sk == 0 else 'f'
                cs, sk, si, so = self.wait([{0: 'l'}, {1: edge}, {3: 'e'}])
            # Save last change.
            packet.append(Packet(self.samplenum, self.matched, cs, sk, si, so))

            # Figure out if this is a status check.
            # Either there is no clock or no start bit (on first rising edge).
            status_check = True
            for change in packet:
                # Get first clock rising edge.
                if len(change.matched) > 1 and change.matched[1] and change.sk:
                    if change.si:
                        status_check = False
                    break

            # The packet is for a status check.
            # SO low = busy, SO high = ready.
            # The SO signal might be noisy in the beginning because it starts
            # in high impedance.
            if status_check:
                start_samplenum = packet[0].samplenum
                bit_so = packet[0].so
                # Check for SO edges.
                for change in packet:
                    if len(change.matched) > 2 and change.matched[2]:
                        if bit_so == 0 and change.so:
                            # Rising edge Busy -> Ready.
                            self.put(start_samplenum, change.samplenum,
                                     self.out_ann, [4, ['Busy', 'B']])
                        start_samplenum = change.samplenum
                        bit_so = change.so
                # Put last state.
                if bit_so == 0:
                    self.put(start_samplenum, packet[-1].samplenum,
                             self.out_ann, [4, ['Busy', 'B']])
                else:
                    self.put(start_samplenum, packet[-1].samplenum,
                             self.out_ann, [3, ['Ready', 'R']])
            else:
                # Bit communication.
                # Since the slave samples SI on clock rising edge we do the
                # same. Because the slave changes SO on clock rising edge we
                # sample on the falling edge.
                bit_start = 0 # Rising clock sample of bit start.
                bit_si = 0 # SI value at rising clock edge.
                bit_so = 0 # SO value at falling clock edge.
                start_bit = True # Start bit incoming (first bit).
                pydata = [] # Python output data.
                for change in packet:
                    if len(change.matched) > 1 and change.matched[1]:
                        # Clock edge.
                        if change.sk: # Rising clock edge.
                            if bit_start > 0: # Bit completed.
                                if start_bit:
                                    if bit_si == 0: # Start bit missing.
                                        self.put(bit_start, change.samplenum,
                                                 self.out_ann,
                                                 [5, ['Start bit not high',
                                                 'Start bit low']])
                                    else:
                                        self.put(bit_start, change.samplenum,
                                                 self.out_ann,
                                                 [0, ['Start bit', 'S']])
                                    start_bit = False
                                else:
                                    self.put(bit_start, change.samplenum,
                                             self.out_ann,
                                             [1, ['SI bit: %d' % bit_si,
                                                  'SI: %d' % bit_si,
                                                  '%d' % bit_si]])
                                    self.put(bit_start, change.samplenum,
                                             self.out_ann,
                                             [2, ['SO bit: %d' % bit_so,
                                                  'SO: %d' % bit_so,
                                                  '%d' % bit_so]])
                                    pydata.append(PyPacket(bit_start,
                                        change.samplenum, bit_si, bit_so))
                            bit_start = change.samplenum
                            bit_si = change.si
                        else: # Falling clock edge.
                            bit_so = change.so
                    elif change.matched[0] and \
                                    change.cs == 0 and change.sk == 0:
                        # End of packet.
                        self.put(bit_start, change.samplenum, self.out_ann,
                                 [1, ['SI bit: %d' % bit_si,
                                      'SI: %d' % bit_si, '%d' % bit_si]])
                        self.put(bit_start, change.samplenum, self.out_ann,
                                 [2, ['SO bit: %d' % bit_so,
                                      'SO: %d' % bit_so, '%d' % bit_so]])
                        pydata.append(PyPacket(bit_start, change.samplenum,
                                      bit_si, bit_so))
                self.put(packet[0].samplenum, packet[len(packet) - 1].samplenum,
                         self.out_python, pydata)



##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2012 Iztok Jeras <iztok.jeras@gmail.com>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
##

import sigrokdecode as srd

# Dictionary of ROM commands and their names, next state.
command = {
    0x33: ['Read ROM'                  , 'GET ROM'   ],
    0x0f: ['Conditional read ROM'      , 'GET ROM'   ],
    0xcc: ['Skip ROM'                  , 'TRANSPORT' ],
    0x55: ['Match ROM'                 , 'GET ROM'   ],
    0xf0: ['Search ROM'                , 'SEARCH ROM'],
    0xec: ['Conditional search ROM'    , 'SEARCH ROM'],
    0x3c: ['Overdrive skip ROM'        , 'TRANSPORT' ],
    0x69: ['Overdrive match ROM'       , 'GET ROM'   ],
    0xa5: ['Resume'                    , 'TRANSPORT' ],
    0x96: ['DS2408: Disable Test Mode' , 'GET ROM'   ],
}

class Decoder(srd.Decoder):
    api_version = 3
    id = 'onewire_network'
    name = '1-Wire network layer'
    longname = '1-Wire serial communication bus (network layer)'
    desc = 'Bidirectional, half-duplex, asynchronous serial bus.'
    license = 'gplv2+'
    inputs = ['onewire_link']
    outputs = ['onewire_network']
    tags = ['Embedded/industrial']
    annotations = (
        ('text', 'Human-readable text'),
    )

    def __init__(self):
        self.reset()

    def reset(self):
        self.ss_block = 0
        self.es_block = 0
        self.state = 'COMMAND'
        self.bit_cnt = 0
        self.search = 'P'
        self.data_p = 0x0
        self.data_n = 0x0
        self.data = 0x0
        self.rom = 0x0000000000000000

    def start(self):
        self.out_python = self.register(srd.OUTPUT_PYTHON)
        self.out_ann = self.register(srd.OUTPUT_ANN)

    def putx(self, data):
        # Helper function for most annotations.
        self.put(self.ss_block, self.es_block, self.out_ann, data)

    def puty(self, data):
        # Helper function for most protocol packets.
        self.put(self.ss_block, self.es_block, self.out_python, data)

    def decode(self, ss, es, data):
        code, val = data

        # State machine.
        if code == 'RESET/PRESENCE':
            self.search = 'P'
            self.bit_cnt = 0
            self.put(ss, es, self.out_ann,
                     [0, ['Reset/presence: %s' % ('true' if val else 'false')]])
            self.put(ss, es, self.out_python, ['RESET/PRESENCE', val])
            self.state = 'COMMAND'
            return

        # For now we're only interested in 'RESET/PRESENCE' and 'BIT' packets.
        if code != 'BIT':
            return

        if self.state == 'COMMAND':
            # Receiving and decoding a ROM command.
            if self.onewire_collect(8, val, ss, es) == 0:
                return
            if self.data in command:
                self.putx([0, ['ROM command: 0x%02x \'%s\''
                          % (self.data, command[self.data][0])]])
                self.state = command[self.data][1]
            else:
                self.putx([0, ['ROM command: 0x%02x \'%s\''
                          % (self.data, 'unrecognized')]])
                self.state = 'COMMAND ERROR'
        elif self.state == 'GET ROM':
            # A 64 bit device address is selected.
            # Family code (1 byte) + serial number (6 bytes) + CRC (1 byte)
            if self.onewire_collect(64, val, ss, es) == 0:
                return
            self.rom = self.data & 0xffffffffffffffff
            self.putx([0, ['ROM: 0x%016x' % self.rom]])
            self.puty(['ROM', self.rom])
            self.state = 'TRANSPORT'
        elif self.state == 'SEARCH ROM':
            # A 64 bit device address is searched for.
            # Family code (1 byte) + serial number (6 bytes) + CRC (1 byte)
            if self.onewire_search(64, val, ss, es) == 0:
                return
            self.rom = self.data & 0xffffffffffffffff
            self.putx([0, ['ROM: 0x%016x' % self.rom]])
            self.puty(['ROM', self.rom])
            self.state = 'TRANSPORT'
        elif self.state == 'TRANSPORT':
            # The transport layer is handled in byte sized units.
            if self.onewire_collect(8, val, ss, es) == 0:
                return
            self.putx([0, ['Data: 0x%02x' % self.data]])
            self.puty(['DATA', self.data])
        elif self.state == 'COMMAND ERROR':
            # Since the command is not recognized, print raw data.
            if self.onewire_collect(8, val, ss, es) == 0:
                return
            self.putx([0, ['ROM error data: 0x%02x' % self.data]])

    # Data collector.
    def onewire_collect(self, length, val, ss, es):
        # Storing the sample this sequence begins with.
        if self.bit_cnt == 0:
            self.ss_block = ss
        self.data = self.data & ~(1 << self.bit_cnt) | (val << self.bit_cnt)
        self.bit_cnt += 1
        # Storing the sample this sequence ends with.
        # In case the full length of the sequence is received, return 1.
        if self.bit_cnt == length:
            self.es_block = es
            self.data = self.data & ((1 << length) - 1)
            self.bit_cnt = 0
            return 1
        else:
            return 0

    # Search collector.
    def onewire_search(self, length, val, ss, es):
        # Storing the sample this sequence begins with.
        if (self.bit_cnt == 0) and (self.search == 'P'):
            self.ss_block = ss

        if self.search == 'P':
            # Master receives an original address bit.
            self.data_p = self.data_p & ~(1 << self.bit_cnt) | \
                          (val << self.bit_cnt)
            self.search = 'N'
        elif self.search == 'N':
            # Master receives a complemented address bit.
            self.data_n = self.data_n & ~(1 << self.bit_cnt) | \
                          (val << self.bit_cnt)
            self.search = 'D'
        elif self.search == 'D':
            # Master transmits an address bit.
            self.data = self.data & ~(1 << self.bit_cnt) | (val << self.bit_cnt)
            self.search = 'P'
            self.bit_cnt += 1

        # Storing the sample this sequence ends with.
        # In case the full length of the sequence is received, return 1.
        if self.bit_cnt == length:
            self.es_block = es
            self.data_p = self.data_p & ((1 << length) - 1)
            self.data_n = self.data_n & ((1 << length) - 1)
            self.data = self.data & ((1 << length) - 1)
            self.search = 'P'
            self.bit_cnt = 0
            return 1
        else:
            return 0



##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2011 Gareth McMullin <gareth@blacksphere.co.nz>
## Copyright (C) 2012-2013 Uwe Hermann <uwe@hermann-uwe.de>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
##

import sigrokdecode as srd

'''
OUTPUT_PYTHON format:

Packet:
[<ptype>, <pdata>]

<ptype>, <pdata>:
 - 'SOP', None
 - 'SYM', <sym>
 - 'BIT', <bit>
 - 'STUFF BIT', None
 - 'EOP', None
 - 'ERR', None
 - 'KEEP ALIVE', None
 - 'RESET', None

<sym>:
 - 'J', 'K', 'SE0', or 'SE1'

<bit>:
 - '0' or '1'
 - Note: Symbols like SE0, SE1, and the J that's part of EOP don't yield 'BIT'.
'''

# Low-/full-speed symbols.
# Note: Low-speed J and K are inverted compared to the full-speed J and K!
symbols = {
    'low-speed': {
        # (<dp>, <dm>): <symbol/state>
        (0, 0): 'SE0',
        (1, 0): 'K',
        (0, 1): 'J',
        (1, 1): 'SE1',
    },
    'full-speed': {
        # (<dp>, <dm>): <symbol/state>
        (0, 0): 'SE0',
        (1, 0): 'J',
        (0, 1): 'K',
        (1, 1): 'SE1',
    },
    'automatic': {
        # (<dp>, <dm>): <symbol/state>
        (0, 0): 'SE0',
        (1, 0): 'FS_J',
        (0, 1): 'LS_J',
        (1, 1): 'SE1',
    },
    # After a PREamble PID, the bus segment between Host and Hub uses LS
    # signalling rate and FS signalling polarity (USB 2.0 spec, 11.8.4: "For
    # both upstream and downstream low-speed data, the hub is responsible for
    # inverting the polarity of the data before transmitting to/from a
    # low-speed port.").
    'low-speed-rp': {
        # (<dp>, <dm>): <symbol/state>
        (0, 0): 'SE0',
        (1, 0): 'J',
        (0, 1): 'K',
        (1, 1): 'SE1',
    },
}

bitrates = {
    'low-speed': 1500000, # 1.5Mb/s (+/- 1.5%)
    'low-speed-rp': 1500000, # 1.5Mb/s (+/- 1.5%)
    'full-speed': 12000000, # 12Mb/s (+/- 0.25%)
    'automatic': None
}

sym_annotation = {
    'J': [0, ['J']],
    'K': [1, ['K']],
    'SE0': [2, ['SE0', '0']],
    'SE1': [3, ['SE1', '1']],
}

class SamplerateError(Exception):
    pass

class Decoder(srd.Decoder):
    api_version = 3
    id = 'usb_signalling'
    name = 'USB signalling'
    longname = 'Universal Serial Bus (LS/FS) signalling'
    desc = 'USB (low-speed/full-speed) signalling protocol.'
    license = 'gplv2+'
    inputs = ['logic']
    outputs = ['usb_signalling']
    tags = ['PC']
    channels = (
        {'id': 'dp', 'name': 'D+', 'desc': 'USB D+ signal'},
        {'id': 'dm', 'name': 'D-', 'desc': 'USB D- signal'},
    )
    options = (
        {'id': 'signalling', 'desc': 'Signalling',
            'default': 'automatic', 'values': ('automatic', 'full-speed', 'low-speed')},
    )
    annotations = (
        ('sym-j', 'J symbol'),
        ('sym-k', 'K symbol'),
        ('sym-se0', 'SE0 symbol'),
        ('sym-se1', 'SE1 symbol'),
        ('sop', 'Start of packet (SOP)'),
        ('eop', 'End of packet (EOP)'),
        ('bit', 'Bit'),
        ('stuffbit', 'Stuff bit'),
        ('error', 'Error'),
        ('keep-alive', 'Low-speed keep-alive'),
        ('reset', 'Reset'),
    )
    annotation_rows = (
        ('bits', 'Bits', (4, 5, 6, 7, 8, 9, 10)),
        ('symbols', 'Symbols', (0, 1, 2, 3)),
    )

    def __init__(self):
        self.reset()

    def reset(self):
        self.samplerate = None
        self.oldsym = 'J' # The "idle" state is J.
        self.ss_block = None
        self.bitrate = None
        self.bitwidth = None
        self.samplepos = None
        self.samplenum_target = None
        self.samplenum_edge = None
        self.samplenum_lastedge = 0
        self.edgepins = None
        self.consecutive_ones = 0
        self.bits = None
        self.state = 'IDLE'

    def start(self):
        self.out_python = self.register(srd.OUTPUT_PYTHON)
        self.out_ann = self.register(srd.OUTPUT_ANN)

    def metadata(self, key, value):
        if key == srd.SRD_CONF_SAMPLERATE:
            self.samplerate = value
            self.signalling = self.options['signalling']
            if self.signalling != 'automatic':
                self.update_bitrate()

    def update_bitrate(self):
        self.bitrate = bitrates[self.signalling]
        self.bitwidth = float(self.samplerate) / float(self.bitrate)

    def putpx(self, data):
        s = self.samplenum_edge
        self.put(s, s, self.out_python, data)

    def putx(self, data):
        s = self.samplenum_edge
        self.put(s, s, self.out_ann, data)

    def putpm(self, data):
        e = self.samplenum_edge
        self.put(self.ss_block, e, self.out_python, data)

    def putm(self, data):
        e = self.samplenum_edge
        self.put(self.ss_block, e, self.out_ann, data)

    def putpb(self, data):
        s, e = self.samplenum_lastedge, self.samplenum_edge
        self.put(s, e, self.out_python, data)

    def putb(self, data):
        s, e = self.samplenum_lastedge, self.samplenum_edge
        self.put(s, e, self.out_ann, data)

    def set_new_target_samplenum(self):
        self.samplepos += self.bitwidth
        self.samplenum_target = int(self.samplepos)
        self.samplenum_lastedge = self.samplenum_edge
        self.samplenum_edge = int(self.samplepos - (self.bitwidth / 2))

    def wait_for_sop(self, sym):
        # Wait for a Start of Packet (SOP), i.e. a J->K symbol change.
        if sym != 'K' or self.oldsym != 'J':
            return
        self.consecutive_ones = 0
        self.bits = ''
        self.update_bitrate()
        self.samplepos = self.samplenum - (self.bitwidth / 2) + 0.5
        self.set_new_target_samplenum()
        self.putpx(['SOP', None])
        self.putx([4, ['SOP', 'S']])
        self.state = 'GET BIT'

    def handle_bit(self, b):
        if self.consecutive_ones == 6:
            if b == '0':
                # Stuff bit.
                self.putpb(['STUFF BIT', None])
                self.putb([7, ['Stuff bit: 0', 'SB: 0', '0']])
                self.consecutive_ones = 0
            else:
                self.putpb(['ERR', None])
                self.putb([8, ['Bit stuff error', 'BS ERR', 'B']])
                self.state = 'IDLE'
        else:
            # Normal bit (not a stuff bit).
            self.putpb(['BIT', b])
            self.putb([6, ['%s' % b]])
            if b == '1':
                self.consecutive_ones += 1
            else:
                self.consecutive_ones = 0

    def get_eop(self, sym):
        # EOP: SE0 for >= 1 bittime (usually 2 bittimes), then J.
        self.set_new_target_samplenum()
        self.putpb(['SYM', sym])
        self.putb(sym_annotation[sym])
        self.oldsym = sym
        if sym == 'SE0':
            pass
        elif sym == 'J':
            # Got an EOP.
            self.putpm(['EOP', None])
            self.putm([5, ['EOP', 'E']])
            self.state = 'WAIT IDLE'
        else:
            self.putpm(['ERR', None])
            self.putm([8, ['EOP Error', 'EErr', 'E']])
            self.state = 'IDLE'

    def get_bit(self, sym):
        self.set_new_target_samplenum()
        b = '0' if self.oldsym != sym else '1'
        self.oldsym = sym
        if sym == 'SE0':
            # Start of an EOP. Change state, save edge
            self.state = 'GET EOP'
            self.ss_block = self.samplenum_lastedge
        else:
            self.handle_bit(b)
        self.putpb(['SYM', sym])
        self.putb(sym_annotation[sym])
        if len(self.bits) <= 16:
            self.bits += b
        if len(self.bits) == 16 and self.bits == '0000000100111100':
            # Sync and low-speed PREamble seen
            self.putpx(['EOP', None])
            self.state = 'IDLE'
            self.signalling = 'low-speed-rp'
            self.update_bitrate()
            self.oldsym = 'J'
        if b == '0':
            edgesym = symbols[self.signalling][tuple(self.edgepins)]
            if edgesym not in ('SE0', 'SE1'):
                if edgesym == sym:
                    self.bitwidth = self.bitwidth - (0.001 * self.bitwidth)
                    self.samplepos = self.samplepos - (0.01 * self.bitwidth)
                else:
                    self.bitwidth = self.bitwidth + (0.001 * self.bitwidth)
                    self.samplepos = self.samplepos + (0.01 * self.bitwidth)

    def handle_idle(self, sym):
        self.samplenum_edge = self.samplenum
        se0_length = float(self.samplenum - self.samplenum_lastedge) / self.samplerate
        if se0_length > 2.5e-6: # 2.5us
            self.putpb(['RESET', None])
            self.putb([10, ['Reset', 'Res', 'R']])
            self.signalling = self.options['signalling']
        elif se0_length > 1.2e-6 and self.signalling == 'low-speed':
            self.putpb(['KEEP ALIVE', None])
            self.putb([9, ['Keep-alive', 'KA', 'A']])

        if sym == 'FS_J':
            self.signalling = 'full-speed'
            self.update_bitrate()
        elif sym == 'LS_J':
            self.signalling = 'low-speed'
            self.update_bitrate()
        self.oldsym = 'J'
        self.state = 'IDLE'

    def decode(self):
        if not self.samplerate:
            raise SamplerateError('Cannot decode without samplerate.')

        # Seed internal state from the very first sample.
        pins = self.wait()
        sym = symbols[self.options['signalling']][pins]
        self.handle_idle(sym)

        while True:
            # State machine.
            if self.state == 'IDLE':
                # Wait for any edge on either DP and/or DM.
                pins = self.wait([{0: 'e'}, {1: 'e'}])
                sym = symbols[self.signalling][pins]
                if sym == 'SE0':
                    self.samplenum_lastedge = self.samplenum
                    self.state = 'WAIT IDLE'
                else:
                    self.wait_for_sop(sym)
                self.edgepins = pins
            elif self.state in ('GET BIT', 'GET EOP'):
                # Wait until we're in the middle of the desired bit.
                self.edgepins = self.wait([{'skip': self.samplenum_edge - self.samplenum}])
                pins = self.wait([{'skip': self.samplenum_target - self.samplenum}])

                sym = symbols[self.signalling][pins]
                if self.state == 'GET BIT':
                    self.get_bit(sym)
                elif self.state == 'GET EOP':
                    self.get_eop(sym)
            elif self.state == 'WAIT IDLE':
                # Skip "all-low" input. Wait for high level on either DP or DM.
                pins = self.wait()
                while not pins[0] and not pins[1]:
                    pins = self.wait([{0: 'h'}, {1: 'h'}])
                if self.samplenum - self.samplenum_lastedge > 1:
                    sym = symbols[self.options['signalling']][pins]
                    self.handle_idle(sym)
                else:
                    sym = symbols[self.signalling][pins]
                    self.wait_for_sop(sym)
                self.edgepins = pins



##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2019 Vesa-Pekka Palmu <vpalmu@depili.fi>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
##

'''
This decoder stacks on top of the 'uart' PD and decodes the ASCII protocol
for Amulet LCD display controllers.

Currently the decoder treats both RX and TX the same way, decoding all
message types.
'''

from .pd import Decoder
